import numpy as np
import numpy.linalg as linalg
import cv2
from skimage import measure
import skimage
from scipy.ndimage import gaussian_filter
import warnings
from skimage.morphology import erosion
from skimage.transform import resize
import mrcfile


warnings.filterwarnings("ignore")


class DataHolder:
    def __init__(self, I, area, stop_condition_n, max_iter, contamination_criterion):
        self.stop_condition_n = stop_condition_n
        self.max_iter = max_iter
        self.area = int(area)
        self.cov_mat_size = area * area
        self.contamination_criterion = contamination_criterion

        # Find shape of I and computing phi_0
        self.size_x, self.size_y = I.shape
        self.phi = initialize_phi_0(self.size_x, self.size_y)
        self.old_phi = self.phi.copy()

        # Defining I_patches
        self.I = I
        self.I_patches = np.empty((self.size_x // area, self.size_y // area, self.cov_mat_size))
        self.I_patches_vector = np.empty(((self.size_x // area) * (self.size_y // area), self.cov_mat_size))
        I_patches = self.I_patches
        for x in range(self.size_x // area):
            for y in range(self.size_y // area):
                I_patches[x, y] = self.I[x * area:(x + 1) * area, y * area:(y + 1) * area].flatten()

        self.I_patches_vector = np.reshape(I_patches, ((self.size_x // area) * (self.size_y // area), self.cov_mat_size))

        # Other variables
        self.iteration = 0
        self.mu0_est = np.empty(self.cov_mat_size)
        self.mu1_est = np.empty(self.cov_mat_size)
        self.cov0_inv = np.empty((self.cov_mat_size, self.cov_mat_size))
        self.cov1_inv = np.empty((self.cov_mat_size, self.cov_mat_size))
        self.logdet0 = 0
        self.logdet1 = 0

    def compute_statistics(self):
        phi_max = skimage.measure.block_reduce(self.phi, (self.area, self.area), np.max)
        phi_min = skimage.measure.block_reduce(self.phi, (self.area, self.area), np.min)

        patch_0 = self.I_patches[np.where(phi_max >= 0)].T
        patch_1 = self.I_patches[np.where(phi_min <= 0)].T

        # Compute mean and covariance
        self.mu0_est = np.mean(patch_0, 1)
        cov0_est = np.cov(patch_0)
        self.mu1_est = np.mean(patch_1, 1)
        cov1_est = np.cov(patch_1)

        self.cov0_inv = linalg.pinv(cov0_est)
        self.cov1_inv = linalg.pinv(cov1_est)
        self.logdet0 = logdet_amitay(cov0_est)
        self.logdet1 = logdet_amitay(cov1_est)

    def get_rts(self):
        x0 = self.I_patches_vector - self.mu0_est[np.newaxis, :]
        x1 = self.I_patches_vector - self.mu1_est[np.newaxis, :]
        tmp_logdet = self.logdet1 - self.logdet0
        tmp_qf = multi_quadratic_form(x1, self.cov1_inv) - multi_quadratic_form(x0, self.cov0_inv)
        rts = (tmp_logdet + tmp_qf) / (2 * self.cov_mat_size)
        return np.reshape(rts, (self.size_x // self.area, self.size_y // self.area))

    def step(self, nu, dt, eps):
        rts = -nu + self.get_rts()

        dt_deltas = dt * delta_eps(self.phi[::self.area, ::self.area], eps)
        dt_deltas *= rts
        for i in range(self.area):
            for j in range(self.area):
                self.phi[i::self.area, j::self.area] += dt_deltas

    def neumann_bound_cond_mod(self, h=1):
        self.phi[[0, -1, 0, -1], [0, 0, -1, -1]] = self.phi[[h, -h - 1, h, -h - 1], [h, h, -h - 1, -h - 1]]
        self.phi[[0, -1], 1:-1] = self.phi[[h, -h - 1], 1:-1]
        self.phi[1:-1, [0, -1]] = self.phi[1:-1, [h, -h - 1]]

    def check_stop_criterion(self, tol):
        if self.iteration % self.stop_condition_n == 0:
            area_new = self.phi > 0
            area_old = self.old_phi > 0
            changed = area_new != area_old
            if np.count_nonzero(changed) / np.count_nonzero(area_old) < tol or self.iteration >= self.max_iter:
                return 1
            self.old_phi = self.phi.copy()
        return 0

    def get_phi(self):
        phi = self.phi.copy()
        I = self.I
        if self.contamination_criterion == 'size':
            def est_func(group):
                return np.count_nonzero(group)
        elif self.contamination_criterion == 'power':
            def est_func(group):
                return np.mean(I[group])
        else:
            raise ValueError('contamination_criterion can only be size or power, got {} instead'.
                             format(self.contamination_criterion))

        group1_score = est_func(phi > 0)
        group2_score = est_func(phi < 0)
        if group1_score >= group2_score:
            phi *= -1
        return phi


def ASOCEM_ver1(micrograph, particle_size, downscale_size, area_size, contamination_criterion, out_size, save_path=None):

    # Require odd area_size, and odd downscale_size such that area_size | downscale_size
    area_size = area_size - 1 if area_size % 2 == 0 else area_size

    downscale_size_max = downscale_size

    while downscale_size_max % area_size != 0 or downscale_size_max % 2 == 0:
        downscale_size_max -= 1

    scalingSz = downscale_size_max / max(micrograph.shape)
    downscale_size_min = np.floor(scalingSz * min(micrograph.shape))

    while downscale_size_min % area_size != 0 or downscale_size_min % 2 == 0:
        downscale_size_min -= 1

    if micrograph.shape[0] >= micrograph.shape[1]:
        size_x, size_y = int(downscale_size_max), int(downscale_size_min)
    else:
        size_x, size_y = int(downscale_size_min), int(downscale_size_max)

    # De-noising filter
    sigma = 1.
    I = gaussian_filter(micrograph, sigma, mode='nearest', truncate=np.ceil(2 * sigma) * sigma)

    # Rescaling
    I = cryo_downsample(I, (size_x, size_y))

    # Executing ASOCEM
    phi = ASOCEM(I, area_size, contamination_criterion)

    # Post processing
    scaling_size = downscale_size / max(micrograph.shape)
    d = max(3, int(np.ceil(scaling_size * particle_size / 8)))
    phi[:d] = 0
    phi[-d:] = 0
    phi[:, :d] = 0
    phi[:, -d:] = 0

    phi_seg = np.zeros(phi.shape)
    se_erod = max(area_size, int(np.ceil(scaling_size * particle_size / 6)))
    phi_erod = erosion((phi > 0).astype('float'), np.ones((se_erod, se_erod)))
    connected_components = measure.label(phi_erod)

    group_threshold = (scaling_size * particle_size) ** 2
    group_ids, group_sizes = np.unique(connected_components, return_counts=True)
    indices = np.argsort(-group_sizes)
    group_ids, group_sizes = group_ids[indices], group_sizes[indices]
    tmp = np.full(connected_components.shape, False)
    for group_id, group_size in zip(group_ids, group_sizes):
        if np.sum(phi_erod[connected_components == group_id]) == 0:
            continue
        tmp = np.logical_or(tmp, connected_components == group_id)
        if group_size > group_threshold:
            phi_seg[connected_components == group_id] = 1

    phi_seg = cv2.dilate(phi_seg, np.ones((se_erod, se_erod)))
    # This resize is the closest I could get to matlab, still minor error
    phi_seg_big = resize(phi_seg, out_size, order=0)

    if save_path is not None:
        save_phi_seg = resize(phi_seg, micrograph.shape, order=0)

        with mrcfile.new(save_path, overwrite=True) as mrc_fh:
            mrc_fh.set_data(save_phi_seg.astype('float32').T)

    return phi_seg_big


def ASOCEM(I, area, contamination_criterion):
    dt = 10 ** 0
    nu = 0
    eps = 1
    tol = 10 ** -3
    max_iter = 300
    area = int(area)

    # Chan Vesse time process
    return chan_vesse_process(I, area, dt, nu, eps, max_iter, tol, contamination_criterion)


def chan_vesse_process(I, area, dt, nu, eps, max_iter, tol, contamination_criterion):
    stop_condition_n = 5
    data_holder = DataHolder(I, area, stop_condition_n, max_iter, contamination_criterion)

    while True:
        # Updating iterations
        data_holder.iteration += 1

        # Computing statistics
        data_holder.compute_statistics()

        # Step
        data_holder.step(nu, dt, eps)

        # Neumann bound
        data_holder.neumann_bound_cond_mod()

        # Stop criteria
        if data_holder.check_stop_criterion(tol):
            break

    return data_holder.get_phi()


def initialize_phi_0(size_x, size_y):
    x, y = np.meshgrid(np.arange(-(size_y // 2), size_y // 2 + 1), np.arange(-(size_x // 2), size_x // 2 + 1))
    phi_0 = (min(size_x, size_y) / 3) ** 2 - (x ** 2 + y ** 2)
    phi_0 /= np.max(np.abs(phi_0))
    return neumann_bound_cond_mod(phi_0)


def delta_eps(t, eps):
    return eps / (np.pi * (eps ** 2 + t ** 2))


def multi_quadratic_form(x, A):
    return np.sum(x @ A * x, -1)


def neumann_bound_cond_mod(f, h=1):
    g = f.copy()
    g[[0, -1, 0, -1], [0, 0, -1, -1]] = g[[h, -h - 1, h, -h - 1], [h, h, -h - 1, -h - 1]]
    g[[0, -1], 1:-1] = g[[h, -h - 1], 1:-1]
    g[1:-1, [0, -1]] = g[1:-1, [h, -h - 1]]
    return g


def logdet_amitay(mat):
    eig_vals, _ = linalg.eig(mat)
    eig_vals[eig_vals < 10e-8] = 1
    return np.sum(np.log(eig_vals))


def cryo_downsample(x, out_shape):
    """
    :param x: ndarray of size (N_1,...N_k)
    :param out_shape: iterable of integers of length k. The value in position i (n_i) is the size we want to cut from
        the center of x in dimension i. If the value of n_i <= 0 or >= N_i then the dimension is left as is.
    :return: out: downsampled x
    """
    dtype_in = x.dtype
    in_shape = np.array(x.shape)
    out_shape = np.array([s if 0 < s < in_shape[i] else in_shape[i] for i, s in enumerate(out_shape)])
    fourier_dims = np.array([i for i, s in enumerate(out_shape) if 0 < s < in_shape[i]])
    size_in = np.prod(in_shape[fourier_dims])
    size_out = np.prod(out_shape[fourier_dims])

    fx = crop(np.fft.fftshift(np.fft.fft2(x, axes=fourier_dims), axes=fourier_dims), out_shape)
    out = np.fft.ifft2(np.fft.ifftshift(fx, axes=fourier_dims), axes=fourier_dims) * (size_out / size_in)
    return out.astype(dtype_in)


def crop(x, out_shape):
    """
    :param x: ndarray of size (N_1,...N_k)
    :param out_shape: iterable of integers of length k. The value in position i is the size we want to cut from the
        center of x in dimension i. If the value is <= 0 then the dimension is left as is
    :return: out: The center of x with size outshape.
    """
    in_shape = np.array(x.shape)
    out_shape = np.array([s if s > 0 else in_shape[i] for i, s in enumerate(out_shape)])
    start_indices = in_shape // 2 - out_shape // 2
    end_indices = start_indices + out_shape
    indexer = tuple([slice(i, j) for (i, j) in zip(start_indices, end_indices)])
    out = x[indexer]
    return out



from pathlib import Path
import sys
import progressbar
import subprocess
import time
import argparse
import os
import re
import random

try:
    import cupy

    num_gpus = cupy.cuda.runtime.getDeviceCount()
except:
    pass


def parse_args(has_cupy):
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input-dir', help='Full path of input directory.')
    parser.add_argument('-o', '--output-dir', help='Full path of output directory.', type=check_dir_exists)
    parser.add_argument('-s', '--particle-size', help='Expected size of particles in pixels.', type=check_positive_int)
    parser.add_argument('-p', '--num-particles',
                        help='Number of particles to pick per micrograph. If set to -1 will pick all particles.',
                        default=-1, type=check_positive_int_or_all)
    parser.add_argument('-n', '--num-noise', help='Number of noise images to pick per micrograph.',
                        default=0, type=check_positive_int_or_zero)
    parser.add_argument('-a', '--use-asocem', action='store_true',
                        help='Choose to use ASOCEM to remove contaminations.', default=False)
    parser.add_argument('--save-asocem', action='store_true', help='Choose to save ASOCEM mrc.', default=False)
    parser.add_argument('--asocem-downsample', help='Use to set asocem downsample size.', default=600, type=check_positive_int)
    parser.add_argument('--asocem-area', help='Use to set asocem covariance area size.', default=5, type=check_positive_odd_int)
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose. Choose this to display number of particles and noise images picked from each micrograph during runtime. Otherwise, you get a simple progress bar.',
                        default=False)
    parser.add_argument('--max-processes',
                        help='Limit the number of concurrent processes to run. -1 to let the program choose.',
                        type=check_positive_int_or_all, default=-1)
    parser.add_argument('--only-do-unfinished',
                        help='Only pick micrographs for which no coordinate file exists in the output directory.',
                        action='store_true', default=False)
    parser.add_argument('--version', help='Check version', action='store_true', default=False)

    if has_cupy:
        parser.add_argument('--no-gpu', action='store_true', help="Don't use GPUs.", default=False)
        parser.add_argument('--gpus',
                            help='Indices of GPUs to be used. Valid indices: 0,...,%d. Enter -1 to use all available GPUS.' % (
                                        num_gpus - 1), default=[-1], nargs='+', type=check_range_gpu)
        args = parser.parse_args()
        if args.gpus == [-1]:
            args.gpus = list(range(num_gpus))
        else:
            args.gpus = [x for x in args.gpus if x in range(num_gpus)]
    else:
        args = parser.parse_args()
        args.no_gpu = 1
        args.gpus = []
    return args


def check_positive_int(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue


def check_positive_int_or_zero(value):
    ivalue = int(value)
    if ivalue < 0:
        raise argparse.ArgumentTypeError("%s is an invalid non-negative int value" % value)
    return ivalue


def check_positive_int_or_all(value):
    ivalue = int(value)
    if ivalue == -1:
        return ivalue
    elif ivalue <= 0 and ivalue != -1:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue


def check_positive_odd_int(value):
    ivalue = int(value)
    if ivalue >= 0:
        if ivalue % 2 == 1:
            return ivalue
    raise argparse.ArgumentTypeError("%s is an invalid positive odd int value" % value)


def check_range_gpu(value):
    ivalue = int(value)
    if ivalue == -1:
        return ivalue
    elif ivalue > num_gpus - 1 or ivalue < 0:
        raise argparse.ArgumentTypeError("%s is not in range 0-%s" % (value, num_gpus - 1))
    return ivalue


def check_dir_exists(value):
    output_dir = Path(value)
    if output_dir.is_file():
        raise argparse.ArgumentTypeError("There is already a file with the name %s." % value)
    elif not output_dir.exists():
        raise argparse.ArgumentTypeError(
            'Directory %s does not exist. Please specify an existing directory.' % output_dir)
    return value


def get_args(has_cupy):
    print("\n")
    while True:
        input_dir = Path(input('Enter full path of micrographs MRC files: '))
        num_files = len(list(input_dir.glob("*.mrc")))
        if num_files > 0:
            print("Found %i MRC files." % len(list(input_dir.glob("*.mrc"))))
            break
        elif not input_dir.is_dir():
            print("%s is not a directory." % input_dir)
        else:
            print("Could not find any files in %s." % input_dir)

    while True:
        particle_size = input('Enter the particle size in pixels: ')
        try:
            particle_size = int(particle_size)
            if particle_size < 1:
                print("Particle size must be a positive integer.")
            else:
                break
        except ValueError:
            print("Particle size must be a positive integer.")

    only_do_unfinished = False
    while True:
        output_path = input('Enter full path of output directory: ')
        output_dir = Path(output_path)
        if output_dir.is_file():
            print("There is already a file with the name you specified. Please specify a directory.")
        elif not output_path:
            print("Please specify a directory.")
        elif output_dir.parent.exists() and not output_dir.exists():
            while True:
                create_dir = input('Output directory does not exist. Create? (Y/N): ')
                if create_dir.strip().lower().startswith('y'):
                    Path.mkdir(output_dir)
                    break
                elif create_dir.strip().lower().startswith('n'):
                    print("OK, aborting...")
                    sys.exit(0)
                else:
                    print("Please choose Y/N.")
            break
        elif not output_dir.parent.exists():
            print('Parent directory %s does not exist. Please specify an existing directory.' % output_dir.parent)
        elif output_dir.is_dir():
            num_finished = check_output_dir(input_dir, output_dir, particle_size)
            if num_finished == 1:
                break
            elif num_finished == 2:
                print(
                    "The directory you specified contains coordinate files for all of the micrographs in the input directory. Aborting...")
                sys.exit()
            else:
                while True:
                    only_do_unfinished = input(
                        "The directory you specified contains coordinate files for some of the micrographs in the input directory. Run only on micrographs which have no coordinate file? (Y/N): ")
                    if only_do_unfinished.strip().lower().startswith('y'):
                        only_do_unfinished = True
                        break
                    elif only_do_unfinished.strip().lower().startswith('n'):
                        print("OK, aborting...")
                        sys.exit(0)
                    else:
                        print("Please choose Y/N.")
                break

    num_particles_to_pick = 0
    while num_particles_to_pick == 0:
        pick_all = input('Pick all particles? (Y/N): ')
        if pick_all.strip().lower().startswith('y'):
            num_particles_to_pick = -1
        elif pick_all.strip().lower().startswith('n'):
            while True:
                num_particles_to_pick = input('How many particles to pick: ')
                try:
                    num_particles_to_pick = int(num_particles_to_pick)
                    if num_particles_to_pick < 1:
                        print("Number of particles to pick must be a positive integer.")
                    else:
                        break
                except ValueError:
                    print("Number of particles to pick must be a positive integer.")
        else:
            print("Please choose Y/N.")

    num_noise_to_pick = -1
    while num_noise_to_pick == -1:
        pick_noise = input('Pick noise images? (Y/N): ')
        if pick_noise.strip().lower().startswith('n'):
            num_noise_to_pick = 0
        elif pick_noise.strip().lower().startswith('y'):
            while True:
                num_noise_to_pick = input('How many noise images to pick: ')
                try:
                    num_noise_to_pick = int(num_noise_to_pick)
                    if num_noise_to_pick < 1:
                        print("Number of noise images to pick must be a positive integer.")
                    else:
                        break
                except ValueError:
                    print("Number of particles to pick must be a positive integer.")
        else:
            print("Please choose Y/N.")

    use_asocem = -1
    while use_asocem == -1:
        pick_asocem = input('Do you want to use ASOCEM for contamination removal? (Y/N):')
        if pick_asocem.strip().lower().startswith('n'):
            use_asocem = 0
        elif pick_asocem.strip().lower().startswith('y'):
            use_asocem = 1
        else:
            print("Please choose Y/N.")

    if use_asocem:
        save_asocem_masks = -1
        while save_asocem_masks == -1:
            save_asocem_input = input('Do you want to save ASOCEM masks? (Y/N):')
            if save_asocem_input.strip().lower().startswith('n'):
                save_asocem_masks = 0
            elif save_asocem_input.strip().lower().startswith('y'):
                save_asocem_masks = 1
            else:
                print("Please choose Y/N.")

        change_params = -1
        while change_params == -1:
            change_params_input = input('Do you want to change ASOCEM default parameters? (Y/N):')
            if change_params_input.strip().lower().startswith('n'):
                change_params = 0
            elif change_params_input.strip().lower().startswith('y'):
                change_params = 1
            else:
                print("Please choose Y/N.")

        if change_params:
            # ASOCEM DS
            asocem_downsample = -1
            while asocem_downsample == -1:
                asocem_downsample_input = input('Enter ASOCEM downsample image size (should be a positive number):')
                try:
                    asocem_downsample = int(asocem_downsample_input)
                    if asocem_downsample < 1:
                        print("Downsample image size must be a positive integer.")
                    else:
                        break
                except ValueError:
                    print("Downsample image size must be a positive integer.")

            # ASOCEM area
            asocem_area = -1
            while asocem_area == -1:
                asocem_area_input = input('Enter ASOCEM covariance area size (should be a positive odd number):')
                try:
                    asocem_area = int(asocem_area_input)
                    if asocem_area < 1:
                        print("Covariance area size should be positive.")
                    elif asocem_area % 2 == 0:
                        print("Covariance area size should be odd.")
                    else:
                        break
                except ValueError:
                    print("Covariance area size must be a positive integer.")

        else:
            asocem_downsample = 600
            asocem_area = 5
    else:
        save_asocem_masks = 0
        asocem_downsample = 600
        asocem_area = 5

    verbose = 0
    while verbose == 0:
        verbose_in = input('Display detailed progress? (Y/N): ')
        if verbose_in.strip().lower().startswith('y'):
            verbose = True
        elif verbose_in.strip().lower().startswith('n'):
            verbose = False
            break
        else:
            print("Please choose Y/N.")

    while True:
        max_processes_in = input('Enter maximum number of concurrent processes (-1 to let the program decide): ')
        try:
            max_processes = int(max_processes_in)
            if max_processes < 1 and max_processes != -1:
                print(
                    "Maximum number of concurrent processes must be a positive integer (except -1 to let the program decide).")
            else:
                break
        except ValueError:
            print(
                "Maximum number of concurrent processes must be a positive integer (except -1 to let the program decide).")

    if has_cupy:
        no_gpu = 0
        gpu_indices = []
        while no_gpu == 0:
            no_gpu_in = input('Use GPU? (Y/N): ')
            if no_gpu_in.strip().lower().startswith('n'):
                no_gpu = 1
            elif no_gpu_in.strip().lower().startswith('y'):
                no_gpu = 0
                while len(gpu_indices) == 0:
                    gpu_indices_in = input(
                        'Which GPUs would you like to use? (Valid indices: 0,...,%d. Enter -1 to use all): ' % (
                                    num_gpus - 1))
                    if gpu_indices_in.strip() == '-1':
                        gpu_indices = list(range(num_gpus))
                        break
                    else:
                        gpu_indices_split = re.split(',| ', gpu_indices_in)
                        for gpu_index in gpu_indices_split:
                            try:
                                gpu_index = int(gpu_index)
                                if gpu_index in range(num_gpus):
                                    gpu_indices.append(gpu_index)
                            except ValueError:
                                pass
                        gpu_indices = list(set(gpu_indices))
                        if gpu_indices:
                            break
                        else:
                            print("Please specify valid GPU indices, separated by whitespaces or commas.")
                break
            else:
                print("Please choose Y/N.")

    else:
        no_gpu = 1
        gpu_indices = []
    return input_dir, output_dir, particle_size, num_particles_to_pick, num_noise_to_pick, use_asocem, save_asocem_masks, asocem_downsample, asocem_area, no_gpu, gpu_indices, verbose, max_processes, only_do_unfinished


def check_output_dir(input_dir, output_dir, particle_size):
    """
    Checks how many coordinate files there are in the output directory with
    names matching the micrographs in the input directory, if any.
    If there are no micrographs in input_dir, return 0.
    If the intersection between coordinate file names and micrograph file
    names is empty, return 1.
    If all micrographs have an output coordinate file, return 2.
    Else, return list of micrograph names which do not have an output file.
    """
    mrcs = [mrc for mrc in input_dir.glob("*.mrc")]
    if mrcs == []:
        return 0
    mrc_names = [mrc.name[:-4] for mrc in mrcs]
    output_particles_box_path = output_dir / ("pickedParticlesParticleSize%i/box" % particle_size)
    output = [f for f in output_particles_box_path.glob("*.box")]
    output_names = [f.name[:-4] for f in output]
    intersection = list(set(output_names) & set(mrc_names))
    if len(intersection) == 0:
        return 1
    elif len(intersection) == len(mrc_names):
        return 2
    else:
        unfinished = list(set(mrc_names) - set(output_names))
        return [mrc for mrc in mrcs if mrc.name[:-4] in unfinished]


def progress_bar(output_dir, num_mrcs):
    """
    Progress bar function that reports the progress of the program, by
    periodically checking how many output files have been written. Shows both
    percentage completed and time elapsed.
    """
    start_time = get_start_time(output_dir.parent.parent)
    num_finished = check_num_finished(output_dir, start_time)
    bar = progressbar.ProgressBar(maxval=num_mrcs,
                                  widgets=["[", progressbar.Timer(), "] ", progressbar.Bar('#', '|', '|'), ' (',
                                           progressbar.Percentage(), ')'])
    bar.start()
    while num_finished < num_mrcs:
        num_finished = check_num_finished(output_dir, start_time)
        bar.update(num_finished)
        time.sleep(1)
    bar.finish()
    print("Finished successfully!")


def check_num_finished(output_dir, start_time):
    finished = [f for f in output_dir.glob("*.star") if os.path.getmtime(str(f)) > start_time]
    num_finished = len(finished)
    return num_finished


def get_start_time(output_dir):
    """
    For some reason time.time() and getmtime of a file do not appear to be
    calculated in the same timezone. So we get the start time by checking the
    modification time of a file we create (and immediately delete). A bit ugly,
    but works.
    """
    fp = output_dir / ('%010x' % random.randrange(16 ** 10))
    with fp.open("w") as f:
        f.write("hi")
    start_time = os.path.getmtime(str(fp))
    os.remove(str(fp))
    return start_time


def check_for_newer_version():
    """
    This function checks whether there is a newer version of kltpicker
    available on PyPI. If there is, it issues a warning.

    """
    name = 'kltpicker2'
    # Use pip to try and install a version of kltpicker which does not exist.
    # In answer, you get all available versions. Find the newest one.
    latest_version = str(
        subprocess.run([sys.executable, '-m', 'pip', 'install', '%s==random' % name], capture_output=True, text=True))
    latest_version = latest_version[latest_version.find('(from versions:') + 15:]
    latest_version = latest_version[:latest_version.find(')')]
    latest_version = latest_version.replace(' ', '').split(',')[-1]

    if latest_version == 'none':  # Got an unexpected response.
        pass
    else:  # Use pip to determine the installed version.
        import pkg_resources  # part of setuptools
        current_version = pkg_resources.require('kltpicker2')[0].version
        if latest_version != current_version:
            print(
                "NOTE: you are running an old version of %s (%s). A newer version (%s) is available, please upgrade." % (
                name, current_version, latest_version))



import numpy as np
from .util import f_trans_2, stdfilter, trig_interpolation, radial_avg, fftcorrelate, trig_interpolation_mat,\
    unique_tol, fftconvolve2d_gpu
from scipy import signal
from .cryo_utils import lgwt, cryo_epsds, cryo_prewhiten, picking_from_scoring_mat, als_find_min

try:
    import cupy as cp
except:
    pass

# Globals:
EPS = 10 ** (-2)  # Convergence term for ALS.
PERCENT_EIG_FUNC = 0.99
NUM_QUAD_NYS = 2 ** 7
NUM_QUAD_KER = 2 ** 7
MAX_FUN = 400


class Micrograph:
    """
    Object that contains all the variables and methods needed for the particle picking.

    ...
    Attributes
    ----------
    micrograph : np.ndarray
        Micrograph after downsampling.
    micrograph_pic : np.ndarray
        Original micrograph data.
    mc_size : tuple
        Size of micrograph after downsampling.
    mg_big_size : tuple
        Size of original micrograph.
    noise_mc : np.ndarray
    approx_clean_psd : np.ndarray
        Approximated clean PSD.
    approx_noise_psd : np.ndarray
        Approximated noise PSD.
    approx_noise_var : float
        Approximate noise variance.
    r : np.ndarray
    stop_par : int
        Flag to stop algorithm if maximal number of iterations in PSD approximation was exceeded.
    psd : np.ndarray
    eig_func : np.ndarray
        Eigenfunctions.
    eig_val : np.ndarray
        Eigenvalues.
    num_of_func : int
        Number of eigenfunctions.
    mrc_name : str
        Name of .mrc file.

    Methods
    -------
    cutoff_filter(self, patch_size)
        Radial bandpass filter.
    estimate_rpsd(self, patch_size, max_iter)
    prewhiten_micrograph(self)
    construct_klt_templates(self, kltpicker)
        Constructing the KLTpicker templates as the eigenfunctions of a given kernel.
    detect_particles(self, kltpicker)
        Construct the scoring matrix and then use the picking_from_scoring_mat function to pick particles and noise
        images.
    picking_from_scoring_mat(log_test_n, mrc_name, kltpicker, mg_big_size)
        Pick particles and noise images from the scoring matrix.
    """

    def __init__(self, micrograph, phi_seg, mc_size, mrc_name, mg_big_size):
        self.micrograph = micrograph
        self.phi_seg = phi_seg
        self.mc_size = mc_size
        self.mg_big_size = mg_big_size
        self.noise_mc = micrograph.copy()
        self.approx_clean_psd = 0
        self.approx_noise_psd = 0
        self.approx_noise_var = 0
        self.r = 0
        self.stop_par = 0
        self.psd = 0
        self.eig_func = 0
        self.eig_val = 0
        self.num_of_func = 0
        self.mrc_name = mrc_name

    def cutoff_filter(self, patch_size):
        """Radial bandpass filter."""
        bandpass1d = signal.firwin(int(patch_size), np.array([0.05, 0.95]), pass_zero=False)
        bandpass2d = f_trans_2(bandpass1d)
        return fftcorrelate(self.noise_mc, bandpass2d)

    def estimate_rpsd(self, patch_size, max_iter):
        """
        Approximate clean and noise radial power spectrum density of 
        micrograph.
        """
        phi_seg = self.phi_seg
        noise_mc = self.cutoff_filter(patch_size)
        micro_size = np.min(noise_mc.shape)
        m = np.floor(micro_size / patch_size)
        M = (m ** 2).astype(int)
        L = int(patch_size)
        s = np.zeros((M, L))
        num_quads = 2 ** 7
        quad, nodes = lgwt(num_quads, -np.pi, np.pi)
        x = np.tile(quad, (num_quads, 1))
        rho_mat = (x ** 2 + x.transpose() ** 2) ** 0.5
        rho_mat = np.where(rho_mat > np.pi, 0, rho_mat)
        rho_samp, idx = np.unique(rho_mat, return_inverse=True)

        # make bins for radial average
        N = 2 * L - 1
        X = np.array([[x * 2 / (N - 1) - 1 for x in range(N)] for x in range(N)])
        Y = X.transpose()
        r = (X ** 2 + Y ** 2) ** 0.5
        dr = 1 / (L - 1)
        rbins = np.linspace(-dr / 2, 1 + dr / 2, L + 1)
        R = (rbins[0:-1] + rbins[1:]) / 2
        bins = []
        for j in range(L - 1):
            bins.append(np.where(np.logical_and(r >= rbins[j], r < rbins[j + 1])))
        bins.append(np.where(np.logical_and(r >= rbins[L - 1], r <= 1)))

        trig_interp_mat = trig_interpolation_mat(R * np.pi, rho_samp)  ####
        samples_idx = np.where(np.zeros((int(patch_size), int(patch_size))) == 0)
        for k in range(M):
            row = np.ceil((k + 1) / m).astype(int)
            col = (k + 1 - (row - 1) * m).astype(int)
            if np.any(phi_seg[(row - 1) * L: row * L, (col - 1) * L: col * L]):
                continue
            noisemc_block = noise_mc[(row - 1) * L:row * L, (col - 1) * L: col * L]
            noisemc_block = noisemc_block - np.mean(noisemc_block)
            psd_block = cryo_epsds(noisemc_block, samples_idx, int(np.floor(0.3 * patch_size)))
            r_block = radial_avg(psd_block, L, bins)
            block_var = np.var(noisemc_block, ddof=1)
            psd_rad = np.abs(np.dot(trig_interp_mat, r_block))
            psd_mat = np.reshape(psd_rad[idx], [num_quads, num_quads])
            var_psd = (1 / (2 * np.pi) ** 2) * np.dot(nodes, np.dot(psd_mat, nodes.transpose()))
            scaling_psd = block_var / var_psd
            r_block = scaling_psd * r_block
            s[k] = r_block

        # Remove all 0 rows
        s = s[np.sum(np.abs(s), 1) != 0]

        # find min arg using ALS:
        approx_clean_psd, approx_noise_psd, alpha, stop_par = als_find_min(s, EPS, max_iter)
        std_mat = stdfilter(noise_mc, patch_size)
        var_mat = std_mat ** 2
        var_mat[phi_seg > 0] = 0
        cut = int((patch_size - 1) / 2 + 1)
        var_mat = var_mat[cut - 1:-cut, cut - 1:-cut]
        var_vec = var_mat.transpose().flatten()
        var_vec.sort()
        var_vec = var_vec[var_vec > 0]
        j = np.floor(0.25 * var_vec.size).astype('int')
        noise_var_approx = np.mean(var_vec[0:j])
        num_of_quad = 2 ** 7
        quad, nodes = lgwt(num_of_quad, -np.pi, np.pi)
        y = np.tile(quad, (num_of_quad, 1))
        x = y.transpose()
        rho_mat = np.sqrt(x ** 2 + y ** 2)
        rho_mat = np.where(rho_mat > np.pi, 0, rho_mat)
        rho_samp, idx = np.unique(rho_mat, return_inverse=True)
        clean_psd_nodes = np.abs(trig_interpolation(R * np.pi, approx_clean_psd, rho_samp))
        noise_psd_nodes = np.abs(trig_interpolation(R * np.pi, approx_noise_psd, rho_samp))
        clean_psd_mat = np.reshape(clean_psd_nodes[idx], (num_of_quad, num_of_quad))
        noise_psd_mat = np.reshape(noise_psd_nodes[idx], (num_of_quad, num_of_quad))
        scaling_psd_approx = (np.linalg.multi_dot([nodes, noise_psd_mat, nodes.transpose()]) - (
                4 * np.pi ** 2) * noise_var_approx) / np.linalg.multi_dot([nodes, clean_psd_mat, nodes.transpose()])
        noise_psd_approx_sigma = approx_noise_psd - scaling_psd_approx * approx_clean_psd
        noise_psd_approx_sigma = noise_psd_approx_sigma.clip(min=0, max=None)
        s_mean = np.mean(s, 0)
        s_mean_psd_nodes = np.abs(trig_interpolation(R * np.pi, s_mean, rho_samp))
        s_mean_psd_mat = np.reshape(s_mean_psd_nodes[idx], (num_of_quad, num_of_quad))
        s_mean_var_psd = (1 / (2 * np.pi) ** 2) * np.linalg.multi_dot([nodes, s_mean_psd_mat, nodes.transpose()])
        clean_var_psd = (1 / (2 * np.pi) ** 2) * np.linalg.multi_dot([nodes, clean_psd_mat, nodes.transpose()])
        clean_var = s_mean_var_psd - noise_var_approx
        approx_scaling = clean_var / clean_var_psd
        self.approx_clean_psd = approx_scaling * approx_clean_psd
        self.approx_noise_psd = noise_psd_approx_sigma
        self.approx_noise_var = noise_var_approx
        self.r = R
        self.stop_par = stop_par

    def prewhiten_micrograph(self, kltpicker=None):
        """
        Prewhiten the micrograph using the noise radial power spectrum density.
        """
        if kltpicker is None:
            r = np.floor((self.mc_size[1] - 1) / 2).astype('int')
            c = np.floor((self.mc_size[0] - 1) / 2).astype('int')
            col = np.arange(-c, c + 1) * np.pi / c
            row = np.arange(-r, r + 1) * np.pi / r
            Row, Col = np.meshgrid(row, col)
            rad_mat = np.sqrt(Col ** 2 + Row ** 2)
            rad_samp, idx = unique_tol(rad_mat.flatten(), 1e-14)
        else:
            rad_samp, idx = kltpicker.rsamp_prewhite, kltpicker.idx_rsamp_prewhite
        rad_samp_tmp = rad_samp[rad_samp <= np.max(self.r * np.pi)]
        noise_psd_nodes = np.abs(trig_interpolation(self.r * np.pi, self.approx_noise_psd, rad_samp_tmp))
        noise_psd_nodes = np.pad(noise_psd_nodes, (0, rad_samp.size - noise_psd_nodes.size), 'constant',
                                 constant_values=noise_psd_nodes[-1])
        noise_psd_mat = np.reshape(noise_psd_nodes[idx], self.mc_size)
        noise_mc_prewhite = cryo_prewhiten(self.micrograph, noise_psd_mat)
        tmp = noise_mc_prewhite[self.phi_seg <= 0]
        mean_tmp = np.mean(tmp)
        tmp -= mean_tmp
        noise_mc_prewhite -= mean_tmp
        noise_mc_prewhite = noise_mc_prewhite / np.linalg.norm(tmp)
        self.noise_mc = noise_mc_prewhite

    def construct_klt_templates(self, kltpicker):
        """Constructing the KLTpicker templates as the eigenfunctions of a given kernel."""
        eig_func_tot = np.zeros((kltpicker.max_order, NUM_QUAD_NYS, NUM_QUAD_NYS))
        eig_val_tot = np.zeros((kltpicker.max_order, NUM_QUAD_NYS))
        sqrt_rr = np.sqrt(kltpicker.r_r)
        d_rho_psd_quad_ker = np.diag(kltpicker.rho) * np.diag(self.psd) * np.diag(kltpicker.quad_ker)
        sqrt_diag_quad_nys = np.sqrt(np.diag(kltpicker.quad_nys))

        for n in range(kltpicker.max_order):
            h_nodes = sqrt_rr * np.dot(kltpicker.j_r_rho[n, :, :], np.dot(d_rho_psd_quad_ker,
                                                                          kltpicker.j_r_rho[n, :, :].transpose()))
            tmp = np.dot(sqrt_diag_quad_nys, np.dot(h_nodes, sqrt_diag_quad_nys.transpose()))
            eig_vals, eig_funcs = np.linalg.eigh(tmp)
            eig_vals = np.real(eig_vals)
            eig_vals = eig_vals[::-1]  # Descending.
            eig_funcs = eig_funcs[:, ::-1]
            eig_vals = np.where(np.abs(eig_vals) < np.spacing(1), 0, eig_vals)
            eig_funcs[:, eig_vals == 0] = 0
            eig_func_tot[n] = eig_funcs
            eig_val_tot[n] = eig_vals

        r_idx = np.arange(0, NUM_QUAD_NYS)
        c_idx = np.arange(0, kltpicker.max_order)
        r_idx = np.tile(r_idx, kltpicker.max_order)
        c_idx = np.tile(c_idx, (NUM_QUAD_NYS, 1))
        eig_val_tot = eig_val_tot.flatten()
        r_idx = r_idx.transpose().flatten()
        c_idx = c_idx.transpose().flatten()
        sort_idx = np.argsort(eig_val_tot)
        sort_idx = sort_idx[::-1]
        eig_val_tot = eig_val_tot[sort_idx]
        r_idx = r_idx[sort_idx]
        c_idx = c_idx[sort_idx]
        sum_of_eig = np.sum(eig_val_tot)
        cum_sum_eig_val = np.cumsum(eig_val_tot / sum_of_eig)
        last_eig_idx = (cum_sum_eig_val > PERCENT_EIG_FUNC).argmax() + 1
        eig_val = np.zeros((2 * last_eig_idx, 1))
        eig_func = np.zeros((2 * last_eig_idx, kltpicker.cosine.shape[1]))
        count = 0
        for i in range(last_eig_idx):
            order = c_idx[i]
            idx_of_eig = r_idx[i]
            h_samp = np.sqrt(kltpicker.rsamp_r) * np.linalg.multi_dot([kltpicker.j_samp[order, :, :],
                                                                       np.diag(
                                                                           kltpicker.rho * self.psd * kltpicker.quad_ker),
                                                                       kltpicker.j_r_rho[order, :, :].transpose()])
            v_correct = (1 / np.sqrt(kltpicker.quad_nys)) * eig_func_tot[order, :, idx_of_eig]
            v_nys = np.dot(h_samp, (kltpicker.quad_nys * v_correct)) / eig_val_tot[i]
            v_nys = np.reshape(v_nys[kltpicker.idx_rsamp], kltpicker.idx_rsamp.shape)
            if not order:
                eig_func[count, :] = (1 / np.sqrt(2 * np.pi)) * v_nys
                eig_val[count, 0] = eig_val_tot[i]
                count += 1
            else:
                eig_func[count, :] = np.sqrt((1 / np.pi)) * v_nys * kltpicker.cosine[order, :]
                eig_val[count, 0] = eig_val_tot[i]
                count += 1
                eig_func[count, :] = np.sqrt((1 / np.pi)) * v_nys * kltpicker.sine[order, :]
                eig_val[count, 0] = eig_val_tot[i]
                count += 1
        eig_val = eig_val[eig_val > 0]
        eig_func = eig_func[0:len(eig_val), :]
        if eig_func.shape[0] < MAX_FUN:
            num_of_fun = eig_func.shape[0]
        else:
            num_of_fun = MAX_FUN
        self.eig_func = eig_func
        self.eig_val = eig_val
        self.num_of_func = num_of_fun

    def detect_particles(self, kltpicker):
        """
        Construct the scoring matrix and then use the picking_from_scoring_mat function to pick particles and noise
        images.
        """
        num_of_patch_row = self.mc_size[0] - kltpicker.patch_size_func + 1
        num_of_patch_col = self.mc_size[1] - kltpicker.patch_size_func + 1
        num_of_func = self.num_of_func
        approx_noise_var = self.approx_noise_var

        eig_func_stat = self.eig_func[0:num_of_func, :]
        eig_val_stat = self.eig_val[0:num_of_func]
        for i in range(num_of_func):
            tmp_func = np.reshape(eig_func_stat[i, :], (kltpicker.patch_size_func, kltpicker.patch_size_func))
            tmp_func[kltpicker.rad_mat > np.floor((kltpicker.patch_size_func - 1) / 2)] = 0
            eig_func_stat[i, :] = tmp_func.flatten()
        eig_func_stat = eig_func_stat.T

        c_p = eig_func_stat @ np.diag(eig_val_stat) @ eig_func_stat.T
        c_p = (c_p + c_p.T) / 2
        e, p = np.linalg.eigh(c_p)
        ind_e = np.argsort(-e)
        p = p[:, ind_e]
        e = e[ind_e]
        e = e[:num_of_func]

        # kappa = np.diag(e) + approx_noise_var * np.eye(num_of_func)
        # kappa_inv1 = np.linalg.inv(kappa)
        # t_mat = (1 / approx_noise_var) * np.eye(num_of_func) - kappa_inv1
        # t_mat = np.diagonal(t_mat)
        # mu = np.linalg.slogdet((1 / approx_noise_var) * kappa)[1]

        # Equivalent but nicer, working with only the diagonal instead of the whole matrix
        kappa = e + approx_noise_var
        kappa_inv = 1 / kappa
        t_mat = 1 / approx_noise_var - kappa_inv
        mu = np.sum(np.log(np.abs(kappa))) - num_of_func * np.log(approx_noise_var)

        p = p[:, :num_of_func]
        t_mat = t_mat[:num_of_func]
        if not kltpicker.no_gpu:
            p = cp.asarray(p)
            t_mat = cp.asarray(t_mat)
            noise_mc = cp.asarray(self.noise_mc)
            log_test_mat = cp.zeros((num_of_patch_row, num_of_patch_col))
            for i in range(self.num_of_func):
                qp_tmp = cp.reshape(p[:, i], (kltpicker.patch_size_func, kltpicker.patch_size_func), order='F')
                qp_tmp = cp.flip(cp.flip(qp_tmp, 0), 1)
                scoreTmp = fftconvolve2d_gpu(noise_mc, qp_tmp)
                log_test_mat = log_test_mat + t_mat[i] * scoreTmp ** 2

            log_test_mat = log_test_mat - mu
            neigh = cp.ones((kltpicker.patch_size_func, kltpicker.patch_size_func))
            log_test_n = cp.asnumpy(fftconvolve2d_gpu(log_test_mat, neigh))
        else:
            log_test_mat = np.zeros((num_of_patch_row, num_of_patch_col))
            for i in range(num_of_func):
                qp_tmp = np.reshape(p[:, i], (kltpicker.patch_size_func, kltpicker.patch_size_func), order='F')
                qp_tmp = np.flip(np.flip(qp_tmp, 0), 1)
                scoreTmp = signal.fftconvolve(self.noise_mc, qp_tmp, 'valid')
                log_test_mat = log_test_mat + t_mat[i] * scoreTmp ** 2

            log_test_mat = log_test_mat - mu
            neigh = np.ones((kltpicker.patch_size_func, kltpicker.patch_size_func))
            log_test_n = signal.fftconvolve(log_test_mat, neigh, 'valid')
        [num_picked_particles, num_picked_noise] = picking_from_scoring_mat(log_test_n, self.phi_seg, self.mrc_name,
                                                                            kltpicker, self.mg_big_size)
        return num_picked_particles, num_picked_noise

    def detect_particles_old(self, kltpicker):
        """
        Construct the scoring matrix and then use the picking_from_scoring_mat function to pick particles and noise
        images.
        """
        num_of_patch_row = self.mc_size[0] - kltpicker.patch_size_func + 1
        num_of_patch_col = self.mc_size[1] - kltpicker.patch_size_func + 1
        num_of_func = self.num_of_func
        approx_noise_var = self.approx_noise_var

        eig_func_stat = self.eig_func[0:num_of_func, :]
        eig_val_stat = self.eig_val[0:num_of_func]
        for i in range(num_of_func):
            tmp_func = np.reshape(eig_func_stat[i, :], (kltpicker.patch_size_func, kltpicker.patch_size_func))
            tmp_func[kltpicker.rad_mat > np.floor((kltpicker.patch_size_func - 1) / 2)] = 0
            eig_func_stat[i, :] = tmp_func.flatten()
        eig_func_stat = eig_func_stat.T

        [q, r] = np.linalg.qr(eig_func_stat, 'reduced')
        kappa = np.linalg.multi_dot([r, np.diag(eig_val_stat), r.T]) + (approx_noise_var * np.eye(num_of_func))
        kappa_inv = np.linalg.inv(kappa)
        t_mat = (1 / approx_noise_var) * np.eye(num_of_func) - kappa_inv
        t_mat = (t_mat + t_mat.T) / 2
        mu = np.linalg.slogdet((1 / approx_noise_var) * kappa)[1]

        [D, P] = np.linalg.eigh(t_mat)
        qp = q @ P
        qp = qp.transpose().copy()

        if not kltpicker.no_gpu:
            log_test_n = compute_log_test_n(self.noise_mc, qp, D, mu)
        else:
            log_test_mat = np.zeros((num_of_patch_row, num_of_patch_col))
            for i in range(num_of_func):
                qp_tmp = np.reshape(qp[i, :], (kltpicker.patch_size_func, kltpicker.patch_size_func), order='F')

                qp_tmp = np.flip(np.flip(qp_tmp, 0), 1)

                scoreTmp = signal.fftconvolve(self.noise_mc, qp_tmp, 'valid')
                log_test_mat = log_test_mat + D[i] * abs(scoreTmp ** 2)

            log_test_mat = log_test_mat - mu
            neigh = np.ones((kltpicker.patch_size_func, kltpicker.patch_size_func))
            log_test_n = signal.fftconvolve(log_test_mat, neigh, 'valid')
        [num_picked_particles, num_picked_noise] = picking_from_scoring_mat(log_test_n, self.phi_seg, self.mrc_name,
                                                                            kltpicker, self.mg_big_size)
        return num_picked_particles, num_picked_noise



import numpy as np
from scipy import signal
from scipy.ndimage import uniform_filter
from scipy.fftpack import fftshift

try:
    import cupy as cp
except:
    pass


def fftcorrelate(image, filt):
    """
    Cross-correlate an image and given filter using FFT.
    
    Parameters
    ----------
    image : numpy.ndarray
        First input.
    filt : numpy.ndarray
        Second input.
    
    Returns
    -------
    result : numpy.ndarray
        A 2-dimensional array containing a subset of the discrete linear 
        cross-correlation of image with filt.
    """
    filt = np.rot90(filt, 2)
    pad_shift = 1 - np.mod(np.array(filt.shape), 2)
    filt_center = np.floor((np.array(filt.shape) + 1) / 2).astype("int")
    pad = np.array(filt.shape) - filt_center
    padded_image = np.zeros((image.shape[0] + 2 * pad[0], image.shape[1] + 2 * pad[1]))
    padded_image[pad[0]: pad[0] + image.shape[0], pad[1]: pad[1] + image.shape[1]] = image
    if np.any(pad_shift):
        padded_image = padded_image[pad_shift[0] - 1: -1, pad_shift[1] - 1: -1]
    result = signal.fftconvolve(padded_image, filt, 'valid')
    return result


def f_trans_2(b):
    """
    2-D FIR filter using frequency transformation.

    Produces the 2-D FIR filter h that corresponds to the 1-D FIR filter b 
    using the McClellan transform.
    """
    # McClellan transformation:
    t = np.array([[1, 2, 1], [2, -4, 2], [1, 2, 1]]) / 8
    n = int((b.size - 1) / 2)
    b = np.flip(b, 0)
    b = fftshift(b)
    b = np.flip(b, 0)
    a = 2 * b[0:n + 1]
    a[0] = a[0] / 2
    # Use Chebyshev polynomials to compute h:
    p0 = 1
    p1 = t
    h = a[1] * p1
    rows = 1
    cols = 1
    h[rows, cols] = h[rows, cols] + a[0] * p0
    p2 = 2 * signal.convolve2d(t, p1)
    p2[2, 2] = p2[2, 2] - p0
    for i in range(2, n + 1):
        rows = p1.shape[0] + 1
        cols = p1.shape[1] + 1
        hh = h
        h = a[i] * p2
        h[1:rows, 1:cols] = h[1:rows, 1:cols] + hh
        p0 = p1
        p1 = p2
        rows += 1
        cols += 1
        p2 = 2 * signal.convolve2d(t, p1)
        p2[2:rows, 2:cols] = p2[2:rows, 2:cols] - p0
    h = np.rot90(h, k=2)
    return h


def radial_avg(z, m, bins):
    """
    Radially average 2-D square matrix z into m bins.

    Computes the average along the radius of a unit circle inscribed in the 
    square matrix z. The average is computed in m bins. 
    The radial average is not computed beyond the unit circle, in the corners
    of the matrix z. The radial average is returned in zr.
    
    Parameters
    ----------
    z : numpy.ndarray
        2-D square matrix.
    m : int
        Number of bins.
    bins : numpy.ndarray
        The bins.

    Returns
    -------
    zr : numpy.ndarray
        Radial average of z.
    R : numpy.ndarray
        Mid-points of the bins.
    """
    zr = np.zeros(m)
    for j in range(m):
        n = bins[j][0].size
        if n:
            zr[j] += np.sum(z[bins[j]]) / n
        else:
            zr[j] = np.nan
    return zr


def stdfilter(a, nhood):
    "Local standard deviation of image."
    c1 = uniform_filter(a, nhood, mode='reflect')
    c2 = uniform_filter(a * a, nhood, mode='reflect')
    return np.sqrt(c2 - c1 * c1) * np.sqrt(nhood ** 2. / (nhood ** 2 - 1))


def trig_interpolation_mat(x, xq):
    """
    Matrix for trigonometric interpolation.
    When calculating an interpolant many times with the same nodes and 
    evaluation points, but with different interpolation values at the nodes,
    it is computationally more efficient to construct the following matrix
    once and calculate the trigonometric interpolant by p = dot(mat, y), where
    y is the vector of interpolation values at the nodes x.

    Parameters
    ----------
    x : numpy.ndarray
        interpolation nodes (vector).
    xq : numpy.ndarray
        evaluation points for the interpolant (vector).

    Returns
    -------
    mat : numpy.ndarray
        p = dot(mat, y) where p is the trigonometric interpolant, and y
        are the interpolation values at x.
    """
    n = x.size
    scale = n * (x[1] - x[0]) / 2
    xs = (x / scale) * np.pi / 2
    xi = (xq / scale) * np.pi / 2
    mat = np.zeros((xq.size, x.size))
    if n % 2:
        for k in range(n):
            a = np.sin(n * (xi - xs[k])) / (n * np.sin(xi - xs[k]))
            a[(xi - xs[k]) == 0] = 1
            mat[:, k] = a
    else:
        for k in range(n):
            a = np.sin(n * (xi - xs[k])) / (n * np.tan(xi - xs[k]))
            a[(xi - xs[k]) == 0] = 1
            mat[:, k] = a
    return mat


def trig_interpolation(x, y, xq):
    """   
    Trigonometric interpolation.
    
    Parameters
    ----------
    x : numpy.ndarray
        Interpolation nodes (vector).
    y : numpy.ndarray
        Interpolation values at nodes (vector).
    xq : numpy.ndarray
        Evaluation points for the interpolant (vector).
        
    Returns
    -------
    p : numpy.ndarray
        Values of the trigonometric interpolant (vector).
    """
    n = x.size
    h = 2 / n
    scale = (x[1] - x[0]) / h
    xs = (x / scale) * np.pi / 2
    xi = (xq / scale) * np.pi / 2
    p = np.zeros(xi.size)
    if n % 2:
        for k in range(n):
            a = np.sin(n * (xi - xs[k])) / (n * np.sin(xi - xs[k]))
            a[(xi - xs[k]) == 0] = 1
            p = p + y[k] * a
    else:
        for k in range(n):
            a = np.sin(n * (xi - xs[k])) / (n * np.tan(xi - xs[k]))
            a[(xi - xs[k]) == 0] = 1
            p = p + y[k] * a
    return p


def fftconvolve2d_gpu(x, y):
    """
    Convolve two 2-dimensional arrays using FFT, utilizing CuPy.
    The convolution is in valid mode, meaning that the output consists only of
    those elements that do not rely on the zero-padding.
    
    Parameters
    ----------
    x : numpy.ndarray
        First input.
    y : numpy.ndarray
        Second input.
    
    Returns
    -------
    z : numpy.ndarray
        A 2-dimensional array containing a subset of the discrete linear 
        convolution of x with y.
    """
    xn, xm = x.shape
    yn, ym = y.shape
    zn = xn + yn - 1
    zm = xm + ym - 1
    z = cp.fft.ifft2(cp.fft.fft2(x, s=(zn, zm)) * cp.fft.fft2(y, s=(zn, zm))).real
    valid_n = xn - yn + 1
    valid_m = xm - ym + 1
    start_n = (zn - valid_n) // 2
    start_m = (zm - valid_m) // 2
    end_n = start_n + valid_n
    end_m = start_m + valid_m
    z = z[start_n:end_n, start_m:end_m]
    return z


def unique_tol(x, tol):
    tol *= np.max(np.abs(x))
    sorted_idx = np.argsort(x)
    sorted_x = x[sorted_idx]
    uniques = []
    pos = np.zeros(len(x), dtype=int)
    curr_idx = 0
    while curr_idx < len(x):
        curr_unique = sorted_x[curr_idx]
        last_index = np.searchsorted(sorted_x, curr_unique + tol)
        pos[sorted_idx[curr_idx:last_index]] = len(uniques)
        uniques.append(curr_unique)
        curr_idx = last_index
    return np.array(uniques), pos



from .ASOCEM_one_sample import ASOCEM_ver1
from pathlib import Path
import warnings
from sys import exit, argv
import numpy as np
from .kltpicker import KLTPicker
from .util import trig_interpolation
from .kltpicker_input import parse_args, get_args, progress_bar, check_num_finished, check_for_newer_version, \
    check_output_dir
import mrcfile
from .micrograph import Micrograph
from .cryo_utils import downsample
import multiprocessing as mp
import os
import time

warnings.filterwarnings("ignore")

# Check if CuPy is installed and we have GPU devices
try:
    import cupy as cp

    cp.cuda.runtime.getDeviceCount()
    HAS_CUPY = 1
except:
    HAS_CUPY = 0


def process_micrograph(micrograph, picker):
    micrograph.estimate_rpsd(picker.patch_size, picker.max_iter)
    micrograph.approx_noise_psd = micrograph.approx_noise_psd + np.median(micrograph.approx_noise_psd) / 10
    micrograph.prewhiten_micrograph(picker)
    micrograph.estimate_rpsd(picker.patch_size, picker.max_iter)
    micrograph.psd = np.abs(trig_interpolation(np.pi * micrograph.r, micrograph.approx_clean_psd,
                                               picker.rho))
    micrograph.construct_klt_templates(picker)
    num_picked_particles, num_picked_noise = micrograph.detect_particles(picker)
    return [micrograph.mrc_name, num_picked_particles, num_picked_noise]


def get_micrograph(mrc_file, picker):
    """Reads an mrc file and downsamples it."""
    mgscale = picker.mgscale
    use_asocem = picker.use_asocem
    mrc = mrcfile.open(mrc_file, permissive=True)
    mrc_data = mrc.data.astype('float64').transpose()
    mrc.close()
    mrc_size = mrc_data.shape
    mrc_data = np.rot90(mrc_data)
    mrc_data_original = mrc_data.copy()
    mrc_data = downsample(mrc_data, (np.floor(np.multiply(mgscale, mrc_size))).astype(int))
    if np.mod(mrc_data.shape[0], 2) == 0:  # Odd size is needed.
        mrc_data = mrc_data[0:-1, :]
    if np.mod(mrc_data.shape[1], 2) == 0:  # Odd size is needed.
        mrc_data = mrc_data[:, 0:-1]

    # Compute phi_seg
    if use_asocem:
        if not picker.output_asocem.exists():
            picker.output_asocem.mkdir()
        save_path = picker.output_asocem / mrc_file.name if picker.save_asocem_masks else None
        phi_seg = ASOCEM_ver1(mrc_data_original, picker.particle_size, picker.asocem_downsample, picker.asocem_area,
                              'size', mrc_data.shape, save_path)
    else:
        phi_seg = np.zeros(mrc_data.shape)
    tmp = mrc_data[phi_seg <= 0]
    mean_tmp = np.mean(tmp)
    tmp -= mean_tmp
    mrc_data -= mean_tmp
    mrc_data = mrc_data / np.linalg.norm(tmp)
    mc_size = mrc_data.shape
    micrograph = Micrograph(mrc_data, phi_seg, mc_size, mrc_file.name, mrc_size)
    return micrograph


def get_mempool_usage(param, gpu_index):
    """
    Calculate how much memory is used in the GPU by a single process on one mrc.

    """
    with cp.cuda.Device(gpu_index):
        mrc_file = param[0]
        picker = param[1]
        start = cp.cuda.runtime.memGetInfo()[0]
        micrograph = get_micrograph(mrc_file, picker)
        process_micrograph(micrograph, picker)
        finish = cp.cuda.runtime.memGetInfo()[0]
        mem_usage = (start - finish) * 1.2
        return mem_usage


def calc_procs_per_gpu(mem_usage, max_processes, gpu_indices):
    """
    Find how many available GPUs there are, and for each available GPU compute
    the maximal number of processes that can use it (based on memory usage).
    """
    num_available_cpus = min(mp.cpu_count(), max_processes)
    procs_per_gpu = {}
    for gpu_index in gpu_indices:
        if sum(procs_per_gpu.values()) < num_available_cpus:
            cp.cuda.runtime.setDevice(gpu_index)
            free_mem = cp.cuda.runtime.memGetInfo()[0]
            procs_per_gpu[gpu_index] = min(np.floor(free_mem / mem_usage),
                                           num_available_cpus - sum(procs_per_gpu.values()))
        else:
            procs_per_gpu[gpu_index] = 0
    return procs_per_gpu


def multi_process_micrograph_gpu(param):
    """
    Process micrographs in parallel, using GPU.
    """
    # Unpack parameters (pool.map can map only one argument per function call).
    mrc_file = param[0]
    picker = param[1]
    gpu_index = param[2]
    with cp.cuda.Device(gpu_index):
        micrograph = get_micrograph(mrc_file, picker)
        summary = process_micrograph(micrograph, picker)
    if picker.verbose:  # User requested detailed output.
        num_finished = check_num_finished(picker.output_particles / 'star', picker.start_time)
        print(time.strftime("%H:%M:%S", time.gmtime(
            time.time() - picker.start_time)) + " - Picked %d particles and %d noise images out of %s. (%3d%s)" % (
              summary[1], summary[2], summary[0], round(num_finished / picker.num_mrcs * 100), "%"))
    return summary


def multi_process_micrograph(param):
    """
    Process micrographs in parallel, no GPU.
    """
    # Unpack parameters (pool.map can map only one argument per function call).
    mrc_file = param[0]
    picker = param[1]
    micrograph = get_micrograph(mrc_file, picker)
    summary = process_micrograph(micrograph, picker)
    if picker.verbose:  # User requested detailed output.
        num_finished = check_num_finished(picker.output_particles / 'star', picker.start_time)
        print(time.strftime("%H:%M:%S", time.gmtime(
            time.time() - picker.start_time)) + " - Picked %d particles and %d noise images out of %s. (%3d%s)" % (
              summary[1], summary[2], summary[0], round(num_finished / picker.num_mrcs * 100), "%"))
    return summary


def multi_process_micrograph_pool(gpu_index, num_procs, batch, shared_list):
    """
    A wrapper function that allows processing many micrographs in parallel using
    a worker pool.
    For each GPU, this function is called in a new process. This process runs
    a pool of workers of size num_procs, all using the same GPU.
    """
    for b in batch:
        b.append(gpu_index)
    with mp.Pool(processes=int(num_procs)) as pool:
        shared_list += [x for x in pool.imap_unordered(multi_process_micrograph_gpu, batch, chunksize=7)]


def get_mrc_batches(params, cpus_per_gpu):
    """
    Split the mrc files into batches proportionally sized to the number of
    processes to be run using each GPU.
    """
    batches = {}
    gpus = sorted(list(cpus_per_gpu.keys()))
    cpus = np.array([cpus_per_gpu[gpu] for gpu in gpus])
    batch_sizes = np.ceil(len(params) * cpus / np.sum(cpus))
    indices = np.zeros(len(batch_sizes) + 1).astype(int)
    indices[1:] = np.cumsum(batch_sizes).astype("int")
    param_batches = [params[indices[i]:indices[i + 1]] for i in range(len(indices) - 1)]
    for i in range(len(param_batches)):
        batches[gpus[i]] = param_batches[i]
    return batches


def preprocess_klt_prewhite(kltpicker, mrc_file):
    # Continue preprocess picker
    mrc = mrcfile.open(mrc_file, permissive=True)
    mrc_data = mrc.data.astype('float64').transpose()
    mrc.close()
    mrc_data = np.rot90(mrc_data)
    mrc_size = np.array(mrc_data.shape)
    kltpicker.preprocess_prewhiten(mrc_size)


def main():
    # Because of CUDA limitations, it is impossible to fork processes after
    # invoking CUDA. So we need to use 'spawn' start method instead.
    mp.set_start_method('spawn', force=True)

    # Get user arguments:
    user_input = argv
    if len(user_input) > 1:  # User entered arguments. Use command line mode.
        args = parse_args(HAS_CUPY)
        # Check if user entered the mandatory arguments: input and output
        # directory and particle size. If not, exit.
        if args.version:
            import pkg_resources  # part of setuptools
            version = pkg_resources.require('kltpicker2')[0].version
            print('kltpicker2 {}'.format(version))
            exit()
        if args.output_dir is None or args.input_dir is None or args.particle_size is None:
            print(
                "Error: one or more of the following arguments are missing: input-dir, output-dir, particle-size. For help run kltpicker -h")
            exit()
        else:
            num_finished_output = check_output_dir(Path(args.input_dir), Path(args.output_dir), args.particle_size)
            if num_finished_output == 2:
                print(
                    "The output directory contains coordinate files for all of the micrographs in the input directory. Aborting...")
                exit()
            elif num_finished_output == 1:
                pass
            elif num_finished_output == 0:
                print("Could not find any .mrc files in %s. \nExiting..." % args.input_dir)
                exit(0)
            else:
                if not args.only_do_unfinished:
                    print(
                        "The output directory contains coordinate files for some of the micrographs in the input directory. Use --only-do-unfinished if needed. Aborting...")
                    exit()

    else:  # User didn't enter arguments, use interactive mode to get arguments.
        args = parse_args(HAS_CUPY)  # Initiate args with default values.
        args.input_dir, args.output_dir, args.particle_size, args.num_particles, args.num_noise, args.use_asocem, args.save_asocem, args.asocem_downsample, args.asocem_area, args.no_gpu, args.gpus, args.verbose, args.max_processes, args.only_do_unfinished = get_args(HAS_CUPY)

    # Check newer version
    try:
        check_for_newer_version()
    except:
        pass

    # Handle user options:
    # If max_processes limit not set, set it to infinity.
    if args.max_processes == -1:
        args.max_processes = np.inf

    # Find number of .mrc files in input directory.
    # Check if output directory already contains any output coordinate files
    # for the micrographs in the input directory. If so, remove these
    # micrographs from the micrographs to be processed.

    mrc_files = check_output_dir(Path(args.input_dir), Path(args.output_dir), args.particle_size)
    if mrc_files == 1:  # Need to process all the micrographs in the input dir.
        mrc_files = list(Path(args.input_dir).glob("*.mrc"))
    print("\nRunning on %i files." % len(mrc_files))

    if not args.no_gpu:
        print("Using GPUs %s." % (", ".join([str(x) for x in args.gpus])))
    if not Path(args.output_dir).exists():  # If the output directory doesn't exist, create it.
        Path.mkdir(args.output_dir)

    picker = KLTPicker(args)  # Initiate picker object.
    picker.num_mrcs = len(mrc_files)

    # Preprocessing. If using GPU, preprocessing includes the calculation of
    # memory that is taken up in the processing of a single micrograph in the
    # GPU.
    print("Preprocessing (usually takes up to 1 minute)...")
    picker.preprocess()
    preprocess_klt_prewhite(picker, mrc_files[0])

    params = [[mrc_file, picker] for mrc_file in mrc_files]

    if args.no_gpu:  # GPU is disabled by user/not available on system.
        print("Preprocess finished. Picking particles...")
        os.environ["NUMBA_DISABLE_CUDA"] = "1"  # Disable use of CUDA by NUMBA.
        if not picker.verbose:  # Display simple progress bar.
            p = mp.Process(target=progress_bar, args=[picker.output_particles / "star", len(mrc_files)],
                           name="KLTPicker_ProgressBar")
            p.start()
            # Pick particles. The number of concurrent processes is the minimum of
        # the limit set by the user and two less than the number of CPUs on the machine.
        with mp.Pool(processes=min(args.max_processes, mp.cpu_count() - 2)) as pool:
            # imap creates an iterator so we don't exhaust the machine's memory (as opposed to map). imap_unordered is slightly faster than imap.
            res = [x for x in pool.imap_unordered(multi_process_micrograph, params)]

    else:  # Using GPU.
        # Calculate the memory usage of the GPU by a single process on one micrograph.
        mem_usage = get_mempool_usage(params[0], args.gpus[0])
        print("Preprocess finished. Picking particles...")

        # Calculate the number of processes to run on each GPU, and partition
        # the micrographs into batches to be passed to each GPU.
        procs_per_gpu = calc_procs_per_gpu(mem_usage, args.max_processes, args.gpus)
        batches = get_mrc_batches(params, procs_per_gpu)
        if not picker.verbose:  # Display simple progress bar.
            p = mp.Process(target=progress_bar, args=[picker.output_particles / "star", len(mrc_files)],
                           name="KLTPicker_ProgressBar")
            p.start()
        # We have multiple processes writing results to the same "res" object,
        # so we need a manager (in the version without GPU the pool function
        # takes care of this).
        manager = mp.Manager()
        res = manager.list()
        # Distribute the batches of micrographs to different processes. Each
        # process runs a pool of workers on its own GPU. The size of each
        # worker pool is according to procs_per_gpu.
        jobs = []
        for i in procs_per_gpu:
            if procs_per_gpu[i]:
                p = mp.Process(target=multi_process_micrograph_pool, args=[i, int(procs_per_gpu[i]), batches[i], res],
                               name="KLTPicker%d" % i)
                jobs.append(p)
                p.start()
        for proc in jobs:
            proc.join()

    # Write summary file and print summary to user.
    num_files = len(res)
    num_particles = sum([row[1] for row in res])
    num_noise = sum([row[2] for row in res])
    print("Picked %d particles and %d noise images out of %d micrographs." % (num_particles, num_noise, num_files))



import numpy as np
from pyfftw.interfaces.numpy_fft import fft2, ifft2
from pyfftw import FFTW
from numpy.polynomial.legendre import leggauss
from numba import jit

try:
    import cupy as cp
except:
    pass


def downsample_gpu(image, size_out):
    """
    Use Fourier methods to change the sample interval and/or aspect ratio
    of any dimensions of the input image. Uses CuPy.
    """
    x = cp.fft.fftshift(cp.fft.fft2(image))
    # crop x:
    nx, ny = x.shape
    nsx = int(cp.floor(nx / 2) - cp.floor(size_out[1] / 2))
    nsy = int(cp.floor(ny / 2) - cp.floor(size_out[0] / 2))
    fx = x[nsx: nsx + size_out[1], nsy: nsy + size_out[0]]
    output = cp.fft.ifft2(cp.fft.ifftshift(fx)) * (cp.prod(cp.array(size_out)) / cp.prod(cp.array(image.shape)))
    return cp.asnumpy(output.real)


def downsample(image, size_out):
    """
    Use Fourier methods to change the sample interval and/or aspect ratio
    of any dimensions of the input image.
    """
    x = np.fft.fftshift(np.fft.fft2(image))
    # crop x:
    nx, ny = x.shape
    nsx = int(np.floor(nx / 2) - np.floor(size_out[1] / 2))
    nsy = int(np.floor(ny / 2) - np.floor(size_out[0] / 2))
    fx = x[nsx: nsx + size_out[1], nsy: nsy + size_out[0]]
    output = np.fft.ifft2(np.fft.ifftshift(fx)) * (np.prod(size_out) / np.prod(image.shape))
    return output.real


def lgwt(n, a, b):
    """
    Get n leggauss points in interval [a, b]
    Parameters
    ----------
    n : int
        Number of points.
    a : float
        Interval starting point.
    b : float
        Interval end point.
    Returns
    -------
    x : numpy.ndarray
        Sample points.
    w : numpy.ndarray
        Weights.
    """
    x1, w = leggauss(n)
    m = (b - a) / 2
    c = (a + b) / 2
    x = m * x1 + c
    w = m * w
    x = np.flipud(x)
    return x, w


def cryo_epsds(image, samples_idx, max_d):
    """
    Estimate the 2D isotropic power spectrum of a given image using only the
    pixels given in samples_idx. Typically, samples_idx will correspond to the
    pixles in the image that are outside a certain radius (where there is no
    particle). The power spectrum is estimated using the correlogram method.

    Parameters
    ----------
    image : numpy.ndarray
        Projection. Must be square and can have odd or even dimensions.
    samples_idx : tuple
        Pixel indices to use for autocorrelation estimation.
    max_d : int
        The autocorrelation is estimated for distances up to max_d. The value
        of max_d should be much less than the number of data samples, N. A
        typical value in the literature is N/1, but this may result in a
        spectrum that is too smooth.

    Returns
    -------
    p2 : numpy.ndarray
        2D power spectrum function. If each image is of size pxp, then P2 is
        of size (2p-1)x(2p-1). P2 is always real.
    r : numpy.ndarray
        1D isotropic autocorrelation function.
    r2 : numpy.ndarray
        2D isotropic autocorrelation function.
    x : numpy.ndarray
        Distances at which the autocorrelction R was estimated.
    """
    p = image.shape[0]
    if max_d >= p:
        max_d = p - 1
    # Estimate the 1D isotropic autocorrelation function.
    r, x, _ = cryo_epsdr(image, samples_idx, max_d)

    # Use the 1D autocorrelation estimated above to populate an array of the 2D
    # isotropic autocorrelction.
    r2 = autocorr_2d(max_d, x, r, p)

    # Window the 2D autocorrelation and Fourier transform it to get the power
    # spectrum. Always use the Gaussian window, as it has positive Fourier
    # transform.
    w = gwindow(p, max_d)

    p2 = np.fft.fftshift(fft2(np.fft.ifftshift(r2 * w))).real

    # Normalize the power spectrum P2. The power spectrum is normalized such
    # that its energy is equal to the average energy of the noise samples used
    # to estimate it.
    e = np.sum(np.square(image[samples_idx] - np.mean(image[samples_idx])))
    mean_e = e / len(samples_idx[0])

    # Normalize p2 such that its mean energy is preserved and is equal to
    # mean_e. That way the mean energy does not go down if the number of
    # pixels is artificially changed (say by upsampling, downsampling, or
    # cropping). Note that p2 is already in units of energy, and so the total
    # energy is given by sum(p2) and not by norm(p2).
    p2 = (p2 / p2.sum()) * mean_e * p2.size

    # Due to the truncation of the Gaussian window, we get small negative
    # values in p2, so we ignore them.
    p2 = np.where(p2 < 0, 0, p2)
    return p2


def cryo_epsdr(image, samples_idx, max_d):
    """
    Estimate the 1D isotropic autocorrelation of an image. The samples to use
    are given in samples_idx. The correlation is computed up to a maximal
    distance of max_d.

    Parameters
    ----------
    image : numpy.ndarray
        square pxp image.
    samples_idx : tuple
        pixel indices to use for autocorrelation estimation.
    max_d : int
        Correlations are computed up to a maximal distance of max_d pixels.
        Default p-1.

    Returns
    -------
    r : numpy.ndarray
        1D vector with samples of the isotropic autocorrelation function.
    x : numpy.ndarray
        Distaces at which the samples of the autocorrelation function are
        given. A vector of the same length as R.
    cnt : numpy.ndarray
        Number of autocorrelation samples available for each distance.

    """
    p = image.shape[0]

    # Generate all possible squared distances. For a vertical shift i and
    # horizontal shift j, dists(i,j) contains the corresponding isotropic
    # correlation distance i^2+j^2. dsquare is then the set of all possible
    # squared distances, that is, all distances that can be generated by
    # integer steps in the horizontal and vertical directions.
    i = np.array([[x for x in range(max_d + 1)] for x in range(max_d + 1)])
    dists = i ** 2 + i.transpose() ** 2
    dsquare = np.sort(np.unique(dists[np.where(dists <= max_d ** 2)]))
    x = dsquare ** 0.5  # Distances at which the correlations are computed.

    # Create a distance map whose value at the index (i,j) is the index in the
    # array dsquare whose value is i^2+j^2. Pairs (i,j) that correspond to
    # distances that are larger than max_d are indicated by (-1).
    dist_map = distmap(max_d, dsquare, dists.shape)
    valid_dists = np.where(dist_map != -1)

    # Compute the number of terms in the expression sum_{j}x(j)x(j+d) for each
    # distance d. As the correlation is two-dimensioanl, we compute for each
    # sum of the form  R(k1,k2)=sum_{i,j} X_{i,j} X_{i+k1,j+k2}, how many
    # summands are in the in it for each (k1,k2). This is done by setting the
    # participating image samples to 1 and computing autocorrelation again.
    mask = np.zeros((p, p))
    mask[samples_idx] = 1
    tmp = np.zeros((2 * p + 1, 2 * p + 1))
    tmp[:p, :p] = mask
    ftmp = fft2(tmp)
    c = ifft2(ftmp * np.conj(ftmp))
    c = c[:max_d + 1, :max_d + 1]
    c = np.round(c.real).astype('int')

    r = np.zeros(len(dsquare))  # r(i) is the value of the ACF at distance x(i)

    # Compute non-periodic autocorrelation of masked image with itself (mask
    # all pixels that are not used to autocorrelation estimation).
    input_fft2 = np.zeros((2 * p + 1, 2 * p + 1), dtype='complex128')
    output_fft2 = np.zeros((2 * p + 1, 2 * p + 1), dtype='complex128')
    input_ifft2 = np.zeros((2 * p + 1, 2 * p + 1), dtype='complex128')
    output_ifft2 = np.zeros((2 * p + 1, 2 * p + 1), dtype='complex128')
    flags = ('FFTW_MEASURE', 'FFTW_UNALIGNED')
    a_fft2 = FFTW(input_fft2, output_fft2, axes=(0, 1), direction='FFTW_FORWARD', flags=flags)
    a_ifft2 = FFTW(input_ifft2, output_ifft2, axes=(0, 1), direction='FFTW_BACKWARD', flags=flags)
    sum_c = c

    input_fft2[samples_idx] = image[samples_idx]
    a_fft2()
    np.multiply(output_fft2, np.conj(output_fft2), out=input_ifft2)
    a_ifft2()
    sum_s = output_ifft2

    # Accumulate all autocorrelation values R(k1,k2) such that k1^2+k2^2=const
    # (all autocorrelations of a certain distance).
    # corrs(i) contains the sum of all products of the form x(j)x(j+d), where
    # d=sqrt(dsquare(i)).
    # corr_count is the number of pairs x(j)x(j+d) for each d.
    corr_count, corrs = accumelate_corrs(len(dsquare), valid_dists, dist_map, sum_c, sum_s)

    # Remove zero correlation sums (distances for which we had no samples at
    # that distance)
    idx = np.where(corr_count != 0)[0]
    r[idx] += corrs[idx] / corr_count[idx]
    cnt = corr_count[idx]
    idx = np.where(corr_count == 0)[0]
    r[idx] = 0
    x[idx] = 0
    return r, x, cnt


@jit(nopython=True)
def accumelate_corrs(dsquare_len, valid_dists, dist_map, sum_c, sum_s):
    """
    Accumulate all autocorrelation values R(k1,k2) such that k1^2+k2^2=const
    (all autocorrelations of a certain distance).
    corrs(i) contains the sum of all products of the form x(j)x(j+d), where
    d=sqrt(dsquare(i)).
    corr_count is the number of pairs x(j)x(j+d) for each d.
    """
    corr_count = np.zeros(dsquare_len)
    corrs = np.zeros(dsquare_len)
    for curr_dist in zip(valid_dists[0], valid_dists[1]):
        dmidx = dist_map[curr_dist]
        corrs[dmidx] += sum_s[curr_dist].real
        corr_count[dmidx] += sum_c[curr_dist]
    return corr_count, corrs


@jit(nopython=True)
def autocorr_2d(max_d, x, r, p):
    """
    Use the 1D autocorrelation r, x to populate an array r2 of the 2D
    isotropic autocorrelation.

    Parameters
    ----------
    max_d : int
        The 1D autocorrelations were computed up to a maximal distance of
        max_d pixels.
    x : numpy.ndarray
        Distaces at which the samples of the autocorrelation function are
        given. A vector of the same length as r.
    r : numpy.ndarray
        1D vector with samples of the isotropic autocorrelation function.
    p : int
        The dimension of the square image that is being processed.

    Returns
    -------
    r2 : numpy.ndarray
        2D isotropic autocorrelation.

    """
    dsquare = x ** 2
    r2 = np.zeros((int(2 * p - 1), int(2 * p - 1)), dtype=np.float64)
    for i in range(-max_d, max_d + 1):
        for j in range(-max_d, max_d + 1):
            d = i ** 2 + j ** 2
            if d <= max_d ** 2:
                idx, _ = bsearch(dsquare, d * (1 - 1e-13), d * (1 + 1e-13))
                r2[i + p - 1, j + p - 1] = r[int(idx) - 1]
    return r2


@jit(nopython=True)
def distmap(max_d, dsquare, dists_shape):
    """
    Create a distance map whose value at the index (i,j) is the index in the
    array dsquare whose value is i^2+j^2. Pairs (i,j) that correspond to
    distances that are larger than max_d are indicated by (-1).
    """
    dist_map = np.zeros(dists_shape)
    for i in range(max_d + 1):
        for j in range(max_d + 1):
            d = i ** 2 + j ** 2
            if d <= max_d ** 2:
                idx, _ = bsearch(dsquare, d - 1e-13, d + 1e-13)
                dist_map[i, j] = idx
    dist_map = dist_map.astype(np.int32) - 1
    return dist_map


@jit(nopython=True)
def gwindow(p, max_d):
    """
    Create 2D Gaussian window for spectral estimation. Return a (2p-1)x(2p-1)
    Gaussian window to be used for 2D power spectrum estimation.

    Parameters
    ----------
    p : int
        Size of the returned mask.
    max_d : int
        Width of the Gaussian.

    Returns
    -------
    w : numpy.ndarray
        (2p-1)x(2p-1) array.
    """
    l = 2 * p - 1
    y = np.array([[x for x in range(l)] for x in range(l)])
    x = y - p + 1
    alpha = float(3)
    # Reciprocal of the standard deviation of the Gaussian window. 1/alpha is
    # the width of the Fourier transform of the window.See Harris 78 for more
    # details.
    w = np.exp(-alpha * (x ** 2 + x.transpose() ** 2) / (2 * max_d ** 2))
    return w


@jit(nopython=True)
def bsearch(x, lower_bound, upper_bound):
    """
    Binary search in a sorted vector.

    Binary O(log2(N)) search of the range of indices of all elements of x
    between LowerBound and UpperBound. If no elements between LowerBound and
    Upperbound are found, the returned lower_index and upper_index are empty.
    The array x is assumed to be sorted from low to high, and is NOT verified
    for such sorting.
    Based on code from
    http://stackoverflow.com/questions/20166847/faster-version-of-find-for-sorted-vectors-matlab

    Parameters
    ----------
    x : numpy.ndarray
        A vector of sorted values from low to high.
    lower_bound : float
        Lower boundary on the values of x in the search.
    upper_bound : flo
        Upper boundary on the values of x in the search.

    Returns
    -------
    lower_idx: int
        The smallest index such that LowerBound<=x(index)<=UpperBound.
    upper_idx: int
        The largest index such that LowerBound<=x(index)<=UpperBound.
    """
    if lower_bound > x[-1] or upper_bound < x[0] or upper_bound < lower_bound:
        return None, None
    lower_idx_a = 1
    lower_idx_b = len(x)
    upper_idx_a = 1
    upper_idx_b = len(x)

    while lower_idx_a + 1 < lower_idx_b or upper_idx_a + 1 < upper_idx_b:
        lw = int(np.floor((lower_idx_a + lower_idx_b) / 2))
        if x[lw - 1] >= lower_bound:
            lower_idx_b = lw
        else:
            lower_idx_a = lw
            if upper_idx_a < lw < upper_idx_b:
                upper_idx_a = lw

        up = int(np.ceil((upper_idx_a + upper_idx_b) / 2))
        if x[up - 1] <= upper_bound:
            upper_idx_a = up
        else:
            upper_idx_b = up
            if lower_idx_a < up < lower_idx_b:
                lower_idx_b = up

    if x[lower_idx_a - 1] >= lower_bound:
        lower_idx = lower_idx_a
    else:
        lower_idx = lower_idx_b
    if x[upper_idx_b - 1] <= upper_bound:
        upper_idx = upper_idx_b
    else:
        upper_idx = upper_idx_a

    if upper_idx < lower_idx:
        return None, None

    return lower_idx, upper_idx

    return lower_idx, upper_idx


def cryo_prewhiten(image, noise_response):
    """
    Pre-whiten a projection using the power spectrum of the noise.

    Parameters
    ----------
    image : numpy.ndarray
        image/projection.
    noise_response : numpy.ndarray
        2d image with the power spectrum of the noise.

    Returns
    -------
    p2 : Pre-whitened image.
    """
    delta = np.finfo(image.dtype).eps

    L1, L2 = image.shape
    l1 = L1 // 2
    l2 = L2 // 2
    K1, K2 = noise_response.shape
    k1 = int(np.ceil(K1 / 2))
    k2 = int(np.ceil(K2 / 2))

    # The whitening filter is the sqrt of of the power spectrum of the noise.
    # Also, normalize the energy of the filter to one.
    filter_var = np.sqrt(noise_response)
    filter_var /= np.linalg.norm(filter_var)

    # The filter should be circularly symmetric. In particular, it is up-down
    # and left-right symmetric. Get rid of any tiny asymmetries in the filter:
    filter_var = (filter_var + np.flipud(filter_var)) / 2
    filter_var = (filter_var + np.fliplr(filter_var)) / 2

    # The filter may have very small values or even zeros. We don't want to
    # process these so make a list of all large entries.
    nzidx = np.where(filter_var > 100 * delta)

    fnz = filter_var[nzidx]
    one_over_fnz = 1 / fnz

    # matrix with 1/fnz in nzidx, 0 elsewhere. Later we multiply the Fourier
    # transform of the padded image with this.
    one_over_fnz_as_mat = np.zeros((noise_response.shape[0], noise_response.shape[1]))
    one_over_fnz_as_mat[nzidx] += one_over_fnz

    # Pad the input projection.
    pp = np.zeros((noise_response.shape[0], noise_response.shape[1]))
    row_start_idx = k1 - l1 - 1
    row_end_idx = k1 + l1
    col_start_idx = k2 - l2 - 1
    col_end_idx = k2 + l2
    if L1 % 2 == 0 and L2 % 2 == 0:
        row_end_idx -= 1
        col_end_idx -= 1
    pp[row_start_idx:row_end_idx, col_start_idx:col_end_idx] = image.copy()

    # Take the Fourier transform of the padded image.
    fp = np.fft.fftshift(np.transpose(fft2(np.transpose(np.fft.ifftshift(pp)))))
    # Divide the image by the whitening filter, but only in places where the
    # filter is large. In frequecies where the filter is tiny  we cannot
    # pre-whiten so we just put zero.
    fp *= one_over_fnz_as_mat
    # pp2 for padded p2.
    pp2 = np.fft.fftshift(np.transpose(ifft2(np.transpose(np.fft.ifftshift(fp)))))
    # The resulting image should be real.
    p2 = np.real(pp2[row_start_idx:row_end_idx, col_start_idx:col_end_idx]).copy()
    return p2


def cryo_prewhiten_gpu(image, noise_response):
    """
    Pre-whiten a projection using the power spectrum of the noise. We accelerate
    the Fourier transform by using GPU (CuPy).

    Parameters
    ----------
    image : numpy.ndarray
        image/projection.
    noise_response : numpy.ndarray
        2d image with the power spectrum of the noise.

    Returns
    -------
    p2 : Pre-whitened image.
    """
    delta = np.finfo(image.dtype).eps

    L1, L2 = image.shape
    l1 = L1 // 2
    l2 = L2 // 2
    K1, K2 = noise_response.shape
    k1 = int(np.ceil(K1 / 2))
    k2 = int(np.ceil(K2 / 2))

    # The whitening filter is the sqrt of of the power spectrum of the noise.
    # Also, normalize the energy of the filter to one.
    filter_var = np.sqrt(noise_response)
    filter_var /= np.linalg.norm(filter_var)

    # The filter should be circularly symmetric. In particular, it is up-down
    # and left-right symmetric. Get rid of any tiny asymmetries in the filter:
    filter_var = (filter_var + np.flipud(filter_var)) / 2
    filter_var = (filter_var + np.fliplr(filter_var)) / 2

    # The filter may have very small values or even zeros. We don't want to
    # process these so make a list of all large entries.
    nzidx = np.where(filter_var > 100 * delta)

    fnz = filter_var[nzidx]
    one_over_fnz = 1 / fnz

    # matrix with 1/fnz in nzidx, 0 elsewhere, later we multiply the Fourier
    # transform of the padded image with this.
    one_over_fnz_as_mat = np.zeros((noise_response.shape[0], noise_response.shape[1]))
    one_over_fnz_as_mat[nzidx] += one_over_fnz

    # Pad the input projection.
    pp = cp.zeros((noise_response.shape[0], noise_response.shape[1]))
    p2 = np.zeros((L1, L2), dtype='complex128')
    row_start_idx = k1 - l1 - 1
    row_end_idx = k1 + l1
    col_start_idx = k2 - l2 - 1
    col_end_idx = k2 + l2
    if L1 % 2 == 0 and L2 % 2 == 0:
        row_end_idx -= 1
        col_end_idx -= 1
    pp[row_start_idx:row_end_idx, col_start_idx:col_end_idx] = cp.asarray(image)

    # Take the Fourier transform of the padded image.
    fp = cp.fft.fftshift(cp.transpose(cp.fft.fft2(cp.transpose(cp.fft.ifftshift(pp)))))
    # Divide the image by the whitening filter, but only in places where the
    # filter is large. In frequecies where the filter is tiny  we cannot
    # pre-whiten so we just put zero.
    fp = fp * cp.asarray(one_over_fnz_as_mat)
    # pp2 for padded p2.
    pp2 = cp.fft.fftshift(cp.transpose(cp.fft.ifft2(cp.transpose(cp.fft.ifftshift(fp)))))
    # The resulting image should be real.
    p2 = cp.real(pp2[row_start_idx:row_end_idx, col_start_idx:col_end_idx])
    return cp.asnumpy(p2)


def als_find_min(sreal, eps, max_iter):
    """
    ALS method for RPSD factorization.

    Approximate Clean and Noise PSD and the particle location vector alpha.
    Parameters
    ----------
    sreal : numpy.ndarray
        PSD matrix to be factorized
    eps : float
        Convergence term
    max_iter : int
        Maximum iterations.

    Returns
    -------
    p2 : numpy.ndarray
        Pre-whitened image.
    approx_clean_psd : numpy.ndarray
        Approximated clean PSD
    approx_noise_psd : numpy.ndarray
        Approximated noise PSD
    alpha_approx : numpy.ndarray
        Particle location vector alpha.
    stop_par : int
        Stop algorithm if an error occurred.
    """
    np.random.seed(1337)
    sreal = sreal.transpose()
    sz = sreal.shape
    patch_num = sz[1]
    One = np.ones(patch_num)
    s_norm_inf = np.apply_along_axis(lambda x: max(np.abs(x)), 0, sreal)
    max_col = np.argmax(s_norm_inf)
    min_col = np.argmin(s_norm_inf)
    clean_sig_tmp = np.abs(sreal[:, max_col] - sreal[:, min_col])
    s_norm_1 = np.apply_along_axis(lambda x: sum(np.abs(x)), 0, sreal)
    min_col = np.argmin(s_norm_1)
    noise_sig_tmp = np.abs(sreal[:, min_col])
    s = sreal - np.outer(noise_sig_tmp, One)
    alpha_tmp = (np.dot(clean_sig_tmp, s)) / np.sum(clean_sig_tmp ** 2)
    alpha_tmp = alpha_tmp.clip(min=0, max=1)
    stop_par = 0
    cnt = 1
    while stop_par == 0:
        if not np.linalg.norm(alpha_tmp, 1):
            alpha_tmp = np.random.random(alpha_tmp.size)
        approx_clean_psd = np.dot(s, alpha_tmp) / sum(alpha_tmp ** 2)
        approx_clean_psd = approx_clean_psd.clip(min=0, max=None)
        s = sreal - np.outer(approx_clean_psd, alpha_tmp)
        approx_noise_psd = np.dot(s, np.ones(patch_num)) / patch_num
        approx_noise_psd = approx_noise_psd.clip(min=0, max=None)
        s = sreal - np.outer(approx_noise_psd, One)
        if not np.linalg.norm(approx_clean_psd, 1):
            approx_clean_psd = np.random.random(approx_clean_psd.size)
        alpha_approx = np.dot(approx_clean_psd, s) / sum(approx_clean_psd ** 2)
        alpha_approx = alpha_approx.clip(min=0, max=1)
        if np.linalg.norm(noise_sig_tmp - approx_noise_psd) / np.linalg.norm(approx_noise_psd) < eps:
            if np.linalg.norm(clean_sig_tmp - approx_clean_psd) / np.linalg.norm(approx_clean_psd) < eps:
                if np.linalg.norm(alpha_approx - alpha_tmp) / np.linalg.norm(alpha_approx) < eps:
                    break
        noise_sig_tmp = approx_noise_psd
        alpha_tmp = alpha_approx
        clean_sig_tmp = approx_clean_psd
        cnt += 1
        if cnt > max_iter:
            stop_par = 1
            break
    return approx_clean_psd, approx_noise_psd, alpha_approx, stop_par


def picking_from_scoring_mat(log_test_n, phi_seg, mrc_name, kltpicker, mg_big_size):
    idx_row = np.arange(log_test_n.shape[0])
    idx_col = np.arange(log_test_n.shape[1])
    [col_idx, row_idx] = np.meshgrid(idx_col, idx_row)
    r_del = np.floor(kltpicker.patch_size_pick_box)
    shape = (log_test_n.shape[1], log_test_n.shape[0])
    log_max = np.max(log_test_n)

    # preparing particle output files:
    box_path = kltpicker.output_particles / 'box'
    star_path = kltpicker.output_particles / 'star'
    if not kltpicker.output_particles.exists():
        kltpicker.output_particles.mkdir()
    if not box_path.exists():
        box_path.mkdir()
    if not star_path.exists():
        star_path.mkdir()
    box_file = open(box_path / mrc_name.replace('.mrc', '.box'), 'w')
    star_file = open(star_path / mrc_name.replace('.mrc', '.star'), 'w')
    star_file.write('data_\n\nloop_\n_rlnCoordinateX #1\n_rlnCoordinateY #2\n_rlnAutopickFigureOfMerit #3\n')
    t = int(kltpicker.patch_size_pick_box) // 2

    # picking particles:
    scoring_mat = log_test_n.copy()
    if kltpicker.num_of_particles == -1:  # pick all particles.
        num_picked = 0
        p_max = kltpicker.threshold + 1
        while num_picked <= kltpicker.max_iter and p_max > kltpicker.threshold:
            max_index = np.argmax(scoring_mat.transpose().flatten())
            p_max = scoring_mat.transpose().flatten()[max_index]
            if not p_max > kltpicker.threshold:
                break
            else:
                [index_col, index_row] = np.unravel_index(max_index, shape)
                ind_row_patch = (index_row - 1) + kltpicker.patch_size_func
                ind_col_patch = (index_col - 1) + kltpicker.patch_size_func
                row_idx_b = row_idx - index_row
                col_idx_b = col_idx - index_col
                rsquare = row_idx_b ** 2 + col_idx_b ** 2
                scoring_mat[rsquare <= (r_del ** 2)] = -np.inf
                ri, rf = max(ind_row_patch - t, 0), min(ind_row_patch + t + 1, phi_seg.shape[0])
                ci, cf = max(ind_col_patch - t, 0), min(ind_col_patch + t + 1, phi_seg.shape[1])
                if np.any(phi_seg[ri: rf, ci: cf]):
                    continue
                box_file.write(
                    '%i\t%i\t%i\t%i\n' % (np.round(
                        (1 / kltpicker.mgscale) * (ind_col_patch + 1 - np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((mg_big_size[1] + 1) - (1 / kltpicker.mgscale) * (
                                                  ind_row_patch + 1 + np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box)))
                star_file.write('%i\t%i\t%f\n' % (
                    np.round((1 / kltpicker.mgscale) * (ind_col_patch + 1)),
                    np.round((mg_big_size[1] + 1) - ((1 / kltpicker.mgscale) * (ind_row_patch + 1))),
                    p_max / log_max))
                num_picked += 1
        star_file.close()
        box_file.close()
        num_picked_particles = num_picked

    else:  # pick only the number of particles specified by user.
        num_picked = 0
        p_max = kltpicker.threshold + 1
        while num_picked <= kltpicker.max_iter and num_picked <= kltpicker.num_of_particles - 1 and p_max > kltpicker.threshold:
            max_index = np.argmax(scoring_mat.transpose().flatten())
            p_max = scoring_mat.transpose().flatten()[max_index]
            if not p_max > kltpicker.threshold:
                break
            else:
                [index_col, index_row] = np.unravel_index(max_index, shape)
                ind_row_patch = (index_row - 1) + kltpicker.patch_size_func
                ind_col_patch = (index_col - 1) + kltpicker.patch_size_func
                row_idx_b = row_idx - index_row
                col_idx_b = col_idx - index_col
                rsquare = row_idx_b ** 2 + col_idx_b ** 2
                scoring_mat[rsquare <= (r_del ** 2)] = -np.inf
                ri, rf = max(ind_row_patch - t, 0), min(ind_row_patch + t + 1, phi_seg.shape[0])
                ci, cf = max(ind_col_patch - t, 0), min(ind_col_patch + t + 1, phi_seg.shape[1])
                if np.any(phi_seg[ri: rf, ci: cf]):
                    continue
                box_file.write(
                    '%i\t%i\t%i\t%i\n' % (np.round(
                        (1 / kltpicker.mgscale) * (ind_col_patch + 1 - np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((mg_big_size[1] + 1) - (1 / kltpicker.mgscale) * (
                                                  ind_row_patch + 1 + np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box)))
                star_file.write('%i\t%i\t%f\n' % (
                    np.round((1 / kltpicker.mgscale) * (ind_col_patch + 1)),
                    np.round((mg_big_size[1] + 1) - ((1 / kltpicker.mgscale) * (ind_row_patch + 1))),
                    p_max / log_max))
                num_picked += 1
        star_file.close()
        box_file.close()
        num_picked_particles = num_picked

    # pick noise:
    if kltpicker.num_of_noise_images != 0:
        scoring_mat = log_test_n.copy()
        p_min = kltpicker.threshold - 1

        # preparing noise output files:
        box_path = kltpicker.output_noise / 'box'
        star_path = kltpicker.output_noise / 'star'
        if not kltpicker.output_noise.exists():
            kltpicker.output_noise.mkdir()
        if not box_path.exists():
            box_path.mkdir()
        if not star_path.exists():
            star_path.mkdir()
        box_file = open(box_path / mrc_name.replace('.mrc', '.box'), 'w')
        star_file = open(star_path / mrc_name.replace('.mrc', '.star'), 'w')
        star_file.write('data_\n\nloop_\n_rlnCoordinateX #1\n_rlnCoordinateY #2\n')

        num_picked = 0
        while num_picked <= kltpicker.num_of_noise_images - 1 and p_min < kltpicker.threshold:
            min_index = np.argmin(scoring_mat.transpose().flatten())
            p_min = scoring_mat.transpose().flatten()[min_index]
            if not p_min < kltpicker.threshold:
                break
            else:
                [index_col, index_row] = np.unravel_index(min_index, shape)
                ind_row_patch = (index_row - 1) + kltpicker.patch_size_func
                ind_col_patch = (index_col - 1) + kltpicker.patch_size_func
                row_idx_b = row_idx - index_row
                col_idx_b = col_idx - index_col
                rsquare = row_idx_b ** 2 + col_idx_b ** 2
                scoring_mat[rsquare <= (r_del ** 2)] = np.inf
                ri, rf = max(ind_row_patch - t, 0), min(ind_row_patch + t + 1, phi_seg.shape[0])
                ci, cf = max(ind_col_patch - t, 0), min(ind_col_patch + t + 1, phi_seg.shape[1])
                if np.any(phi_seg[ri: rf, ci: cf]):
                    continue
                box_file.write(
                    '%i\t%i\t%i\t%i\n' % (np.round(
                        (1 / kltpicker.mgscale) * (ind_col_patch + 1 - np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((mg_big_size[1] + 1) - (1 / kltpicker.mgscale) * (
                                                  ind_row_patch + 1 + np.floor(kltpicker.patch_size_pick_box / 2))),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box),
                                          np.round((1 / kltpicker.mgscale) * kltpicker.patch_size_pick_box)))
                star_file.write('%i\t%i\n' % (np.round((1 / kltpicker.mgscale) * (ind_col_patch + 1)), np.round(
                    (mg_big_size[1] + 1) - ((1 / kltpicker.mgscale) * (ind_row_patch + 1)))))
                num_picked += 1
        star_file.close()
        box_file.close()
        num_picked_noise = num_picked

    else:
        num_picked_noise = 0

    return num_picked_particles, num_picked_noise





from pathlib import Path
import numpy as np
import scipy.special as ssp
from .cryo_utils import lgwt
from .util import unique_tol
from .kltpicker_input import get_start_time
from multiprocessing import Pool, cpu_count

# Globals:
EPS = 10 ** (-2)  # Convergence term for ALS.
PERCENT_EIG_FUNC = 0.99
NUM_QUAD_NYS = 2 ** 7
NUM_QUAD_KER = 2 ** 7
MAX_FUN = 400
MAX_ITER = 6 * (10 ** 4)
MAX_ORDER = 100
THRESHOLD = 0

class KLTPicker:
    """
    KLTpicker object that holds all variables that are used in the computations.

    ...
    Attributes
    ----------
    particle_size : float
        Size of particles to look for in micrographs.
    input_dir : str
        Directory from which to read .mrc files.
    output_dir : str
        Output directory in which to write results.
    no_gpu : bool
        Optional - whether to use GPU or not.
    mgscale : float
        Scaling parameter.
    max_order : int
        Maximal order of eigenfunctions.
    micrographs : np.ndarray
        Array of 2-D micrographs.
    patch_size_pick_box : int
        Particle box size to use.
    num_of_particles : int
        Number of particles to pick per micrograph.
    num_of_noise_images : int
        Number of noise images.
    threshold : float
        Threshold for the picking.
    patch_size : int
        Approximate size of particle after downsampling.
    patch_size_func : int
        Size of disc for computing the eigenfunctions.
    max_iter : int
        Maximal number of iterations for PSD approximation.
    rsamp_length : int
    rad_mat : np.ndarray
    quad_ker : np.ndarray
    quad_nys : np.ndarray
    rho : np.ndarray
    j_r_rho : np.ndarray
    j_samp : np.ndarray
    cosine : np.ndarray
    sine : np.ndarray
    rsamp_r : np.ndarray
    r_r : np.ndarray


    Methods
    -------
    preprocess()
        Initializes parameters needed for the computation.
    get_micrographs()
        Reads .mrc files, downsamples them and adds them to the KLTpicker object.
    """

    def __init__(self, args):
        self.particle_size = args.particle_size
        self.input_dir = Path(args.input_dir)
        self.output_dir = Path(args.output_dir)
        self.output_noise = self.output_dir / ('pickedNoiseParticleSize%d' % args.particle_size)
        self.output_particles = self.output_dir / ('pickedParticlesParticleSize%d' % args.particle_size)
        self.output_asocem = self.output_dir / ('AsocemMasksSize%d' % args.particle_size)
        self.no_gpu = args.no_gpu
        self.use_asocem = args.use_asocem
        self.asocem_downsample = args.asocem_downsample
        self.asocem_area = args.asocem_area
        self.save_asocem_masks = args.save_asocem
        self.mgscale = 101 / args.particle_size
        self.max_order = MAX_ORDER
        self.quad_ker = 0
        self.quad_nys = 0
        self.rho = 0
        self.j_r_rho = 0
        self.j_samp = 0
        self.cosine = 0
        self.sine = 0
        self.rsamp_r = 0
        self.r_r = 0
        self.patch_size_pick_box = np.floor(self.mgscale * args.particle_size)
        self.num_of_particles = args.num_particles
        self.num_of_noise_images = args.num_noise
        self.threshold = THRESHOLD
        patch_size = np.floor(0.8 * self.mgscale * args.particle_size)
        if np.mod(patch_size, 2) == 0:
            patch_size -= 1
        self.patch_size = patch_size
        patch_size_function = np.floor(0.4 * self.mgscale * args.particle_size)
        if np.mod(patch_size_function, 2) == 0:
            patch_size_function -= 1
        self.patch_size_func = int(patch_size_function)
        self.max_iter = MAX_ITER
        self.rsamp_length = 0
        self.rad_mat = 0
        self.idx_rsamp = 0
        self.rad_mat_prewhite = 0
        self.idx_rsamp_prewhite = 0
        self.rsamp_prewhite = 0
        self.verbose = args.verbose
        self.num_mrcs = 0
        self.start_time = get_start_time(self.output_dir)
        print(self.output_dir)

    def preprocess(self):
        """Initializes parameters."""
        radmax = np.floor((self.patch_size_func - 1) / 2)
        x = np.arange(-radmax, radmax + 1, 1)
        X, Y = np.meshgrid(x, x)
        rad_mat = np.sqrt(np.square(X) + np.square(Y))
        rsamp, idx_rsamp = unique_tol(rad_mat.flatten('F'), 1e-14)
        theta = np.arctan2(Y, X).transpose().flatten()
        rho, quad_ker = lgwt(NUM_QUAD_KER, 0, np.pi)
        rho = np.flipud(rho)
        quad_ker = np.flipud(quad_ker)
        r, quad_nys = lgwt(NUM_QUAD_NYS, 0, radmax)
        r = np.flipud(r)
        quad_nys = np.flipud(quad_nys)
        r_r = np.outer(r, r)
        r_rho = np.outer(r, rho)
        rsamp_r = np.outer(np.ones(len(rsamp)), r)
        rsamp_rho = np.outer(rsamp, rho)
        pool = Pool(max(cpu_count() - 2, 1))
        res_j_r_rho = pool.starmap(ssp.jv, [(n, r_rho) for n in range(self.max_order)])
        res_j_samp = pool.starmap(ssp.jv, [(n, rsamp_rho) for n in range(self.max_order)])
        pool.close()
        pool.join()
        j_r_rho = np.squeeze(res_j_r_rho)
        j_samp = np.squeeze(res_j_samp)
        n_times_theta = np.outer(np.arange(self.max_order), theta)
        cosine = np.cos(n_times_theta)
        sine = np.sin(n_times_theta)
        cosine[0, :] = 0
        self.quad_ker = quad_ker
        self.quad_nys = quad_nys
        self.rho = rho
        self.j_r_rho = j_r_rho
        self.j_samp = j_samp
        self.cosine = cosine
        self.sine = sine
        self.rsamp_r = rsamp_r
        self.r_r = r_r
        self.rad_mat = rad_mat
        self.idx_rsamp = idx_rsamp

    def preprocess_prewhiten(self, full_mc_shape):
        """Initializes parameters."""
        new_mc_size = np.floor(self.mgscale * full_mc_shape).astype(int)
        if new_mc_size[0] % 2 == 0:  # Odd size is needed.
            new_mc_size[0] -= 1
        if new_mc_size[1] % 2 == 0:  # Odd size is needed.
            new_mc_size[1] -= 1
        r = np.floor((new_mc_size[1] - 1) / 2).astype('int')
        c = np.floor((new_mc_size[0] - 1) / 2).astype('int')
        col = np.arange(-c, c + 1) * np.pi / c
        row = np.arange(-r, r + 1) * np.pi / r
        Row, Col = np.meshgrid(row, col)
        self.rad_mat_prewhite = np.sqrt(Col ** 2 + Row ** 2)
        self.rsamp_prewhite, self.idx_rsamp_prewhite = unique_tol(self.rad_mat_prewhite.flatten(), 1e-14)



from setuptools import setup, find_packages
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='kltpicker2',
      version='2.0.1',
      description='KLT Picker: Particle picking using data-driven optimal templates',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/ShkolniskyLab/kltpicker2',
      author='Yoel Shkolnisky',
      author_email='yoelsh@tauex.tau.ac.il',
      packages=find_packages(),
      license='GNU General Public License v3.0',
      entry_points={
        "console_scripts": ['kltpicker = kltpicker.main:main']
        },
      install_requires=[
          'numpy>=1.16',
          'mrcfile',
          'scipy>=1.3',
          'pyfftw',
          'progressbar2',
          'numba',
          'scikit-image',
          'opencv-python'
      ],
      extras_require={'gpu': ['cupy']},
      python_requires='~=3.6',
      zip_safe=False)
